package caida.otter;
import java.awt.*;
import caida.tools.GridPanel;

/*****************************************************************************
* File: DimensionDailog.java	Name:DimensionDailog
* Goal: To create a FilenameDialog that works for URLs.
* 	It list the contains of a url and the allow the user to
* 	select a given file out of those in the URL of a dir.
*
* Written: Bradley Huffaker
* For: Cooperative Association for Internet Data Analysis
******************************************************************************
******************************************************************************
*  By accessing this software, DIMENSIONDIALOG, you are duly informed of
*  and agree to be bound by the conditions described below in
*  this notice:
*
*  This software product, DIMENSIONDIALOG, is developed by Bradley Huffaker, 
*  and copyrighted(C) 1998 by the University of California, San Diego (UCSD),
*  with all rights reserved.  UCSD administers the NSF grant to CAIDA,
*  number NCR-9711092, under which this code was developed.
*
*  DIMENSIONDIALOG is a free software. You can redistribute it and/or modify it
*  under the terms of the GNU Public License v.2 as
*  published by UCSD and is incorporated by reference herein.  DIMENSIONDIALOG 
*  is distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF MERCHANTABILITY
*  OR FITNESS FOR A PARTICULAR PURPOSE or that the use of it will not
*  infringe on any third party's intellectual property rights.
*
*  You should have received a copy of the CAIDA General Public
*  License v.1 along with the DIMENSIONDIALOG program.  If for any reason you
*  have not received a copy, please write to
*
*                       University of California, San Diego
*                       SDSC/CAIDA
*                       9500 Gilman Dr., MS-0505
*                       La Jolla, CA 92093 - 0505  USA
*
*  Or contact INFO@CAIDA.ORG
*****************************************************************************/ 
public class DimensionDialog extends Frame
{
	// Start width and height
	// if canceled image will be reset to this
	Dimension dim_default;
	Dimension dim_current;
	Dimension dim_new;
	
	// Input Field
	TextField width_field = new TextField();
	TextField height_field = new TextField();

	// The controling buttons
	// Ok - will close the window and apply any changes made to the canvas
	//	size
	// Apply - will set the current width and height to the canvas
	// Cancel - will set the canvas back to the default width and height

	Button Ok = new Button("Ok");
	Button Apply = new Button("Apply");
	Button Cancel = new Button("Cancel");

	// This is the display that will be changed
	Display display;

	public DimensionDialog(Display display_input)
	{
		super("Set canvas dimensions");

		display = display_input;
		
		GridPanel panel = new GridPanel(this);
		panel.add(new Label("Width"),0,0);	
		panel.add(new Label("Hieght"),0,1);	

		panel.addSize(width_field,1,0,2,1);
		panel.addSize(height_field,1,1,2,1);

		panel.add(Ok,0,2);
		panel.add(Apply,1,2);
		panel.add(Cancel,2,2);
		pack();
	}

	public void show()
	{
		dim_default = display.getImageSize();
		dim_current = new Dimension(dim_default);
		
		width_field.setText(String.valueOf(dim_current.width));
		height_field.setText(String.valueOf(dim_current.height));

		super.show();
		toFront();
	}

	public boolean handleEvent(Event event)
	{
		if (event.id == Event.ACTION_EVENT)
		{
			if (event.target == Ok)
				Ok();
			else if (event.target == Apply)
				Apply();
			else if (event.target == Cancel)
				Cancel();
		}
		return super.handleEvent(event);
	}

	private void Ok()
	{
		Apply();
		hide();
	}

	private void Apply()
	{
		dim_current.width = ParseInt(dim_current.width,width_field);
		dim_current.height = ParseInt(dim_current.height,height_field);
		display.setImageSize(dim_current.width,dim_current.height);
	}

	private int ParseInt(int current_value, TextField field)
	{
		int new_value = current_value;
		try {
			new_value  = Integer.parseInt(field.getText());
		} catch (NumberFormatException error) {
		
			getToolkit().beep();
			field.setText(String.valueOf(current_value));
			new_value = current_value;
		}
		return new_value;
	}

	private void Cancel()
	{
		display.setImageSize(dim_default.width,dim_default.height);
		hide();
	}
}
