package caida.otter;
//import caida.image.WorldProducer;
//import caida.image.WorldProducerInterface;
import caida.tools.PsGenerator;
import java.net.*;
import java.awt.image.FilteredImageSource;
import java.io.*;
import caida.tools.Strings;
import caida.tools.StatusBar;
import java.awt.*;
import java.util.Date;
import java.lang.Math;


/**************************************************************************
* File: Display.java	Name: Display
* Classes: Display,Hover_Check
* Class: Display
* Goal: Create a extended panel which will allow for display
*  and editing of nodes and lines. It will also displays messages
*  to the user thought a status bar which will appear on the 
*  bottom of the display.
*
* Mode:  effects how Display handles events and what message
*  is displayed in the status bar
*	ADD_NODE: will add nodes on mouse down events
*	ADD_LINK: will start a line on mouse down and finish it
*	 on mouse up
*	SELECT: will the nearest object that is with in it size
*	PROPERT: Not current implemented
*	REMOVE: will remove the nearest object
*	FORMAT: set when all events are to be ignored
*	ROOT_SELECT: Not current implemented
*		Should allow the user to select a parent and all 
*		nodes which are its children
*	ZOOM: will zoom with centers the points mouse down event is occured 
*
*	and links.  It will also allow a GUI for thier
*	creation, manipulation, and destruction
*
* Class: Hover_Check
* Goal: To have a class which checks the x, y ever so often and 
* checks to see if the is a object near it and tells display to
* print it.  This means that it must be a thread.	
*    
* Written: Bradley Huffaker             (12/17/97)
*
* For:Cooperative Association for Internet Data Analysis
***************************************************************************
***************************************************************************
By accessing this software, DISPLAY, you are duly informed of and 
agree to be bound by the conditions described below in this notice:
 
This software product, DISPLAY, is developed by Bradley Huffaker,
and copyrighted(C) 1998 by the University of California, San Diego (UCSD), 
with all rights reserved.  UCSD administers the NSF grant to CAIDA,
number NCR-9711092, under which this code was developed."
 
There is no charge for DISPLAY software. You can redistribute it and/or
modify it under the terms of the GNU General Public License, v. 2 dated
June 1991 which is incorporated by reference herein.  DISPLAY is
distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that the use of
it will not infringe on any third party's intellectual property
rights.
 
You should have received a copy of the GNU GPL along with the DISPLAY
program.  Copies can also be obtained from
http://www.gnu.org/copyleft/gpl.html or by writing to
 
                        University of California, San Diego
                        SDSC/CAIDA
                        9500 Gilman Dr., MS-0505
                        La Jolla, CA 92093 - 0505  USA
 
  Or contact INFO@CAIDA.ORG
**************************************************************************/    

//public class Display extends Panel implements Runnable
public class Display extends Panel implements Runnable
{

	// Used to display the current status and other messages
	protected StatusBar statusBar;

	// Flag for when to redraw the whole image or use the old copy
	protected boolean image_changed;	
	// Flag to indicate redrawing the whole map when size is changed 
	protected boolean size_changed;

	// Used to hold the offscreen image
	public Image offscreen;
	protected int offscreen_height = 0;
	protected int offscreen_width = 0;

	// The Rectangle that contains all nodes
	protected int node_height = 0;
	protected int node_width = 0;

	protected int width, height;

	// Map stuff - jjung
	protected double top_lat, top_lon; 
	protected double bot_lat, bot_lon; 
	protected double source_top_lat, source_top_lon; 
	protected double source_bot_lat, source_bot_lon; 
	
	protected Image image;
	protected Image source_image;
	protected Image default_image;
	protected boolean showMap = false;
	protected boolean viewGeo = false;
	protected boolean setMap = false;
	protected boolean stepBystep = false;
	//WorldProducer worldProducer;

	protected int current_step;

	final double MAX_ZOOM = .003;
	double scaler;

	// List of nodes and links
	// List of selected nodes and links
	protected DList dlist;
	protected DList selected;
	protected DList ordered; 
	protected DisplayObject mouse_down_selected = null;
	protected boolean mouse_dragged = false;

	// The date of the file
	protected Date date = null;
	protected boolean date_showen = false;

        public boolean ready = false;

	// Holds the mode that the display is current in
	protected int mode;
	static final public int ADD_NODE = 	0;
 	static final public int ADD_LINK =	1;
 	static final public int SELECT = 	2;
	static final public int ROOT_SELECT = 	3;
	static final public int SELECT_TREE =   4;
 	static final public int PROPERT =	5;
 	static final public int REMOVE =	6;
	static final public int FORMAT = 	7;
	static final public int MOVE_NAMES= 	8;
	static final public int ANIMATE =	9;
	static final public int ZOOM =		10;
	static final public int STEP_BY_STEP =		11;
	static final public int MOVE= 		12;
	static final public int HIDE_SELECT = 	13;
 	static final public String ADD_NODE_STR = 	"Add Node"; 
	static final public String ADD_LINK_STR = 	"Add Link";
 	static final public String SELECT_STR = 	"Select Object";
 	static final public String HIDE_SELECT_STR = 	"Hide Select";
	static final public String ROOT_SELECT_STR =	"Select Root";
	static final public String SELECT_TREE_STR =	"Select Tree";
 	static final public String PROPERT_STR = 	"Properties";
 	static final public String REMOVE_STR = 	"Remove Object";
	static final public String FORMAT_STR =		"Formatting Nodes";
	static final public String MOVE_NAMES_STR =	"Move Names";
	static final public String MOVE_STR =		"Move";
	static final public String ANIMATE_STR =	"None Select";
	static final public String ZOOM_STR =		"Zoom";
	static final public String STEP_BY_STEP_STR =		"Step by Step";

	// Holds the current resolution of the objects
	protected int res;
 	static final public int LOW = 0;
	static final public int HIGH = 1;
	static final public String LOW_STRING = "Low Res";
	static final public String HIGH_STRING = "High Res";

	// Old x and Old y are used in a mouse drag to find the change
	// and move all the selected objects by that change
	int old_x, old_y;

	// Used in the handleEvent to tell if the user is hovering over
	// a object rather just scanning to it.
	int MIN_MOUSE_HOVER = 10;
	
	// This is used by the addlink methods to add a link between
	// two nodes
	Node addlink_node;
	Link addlink_link;

        // This counter is used to know if nodes need to arrange their links
        private int paintCounter = 0;

	/*************************************
	 For some reason I can't make the scroll bars work so I don't have 
	 them
	// This should allow for scroll 
	double scrollbar_x = .5;
	double scrollbar_y = .5;
	***************************************/
	
	// This allows for mouse mostion 
	int x_shift = 0;
	int y_shift = 0;

	// This limits the number to times it checks for closeness
	Check_Hover check_hover;
	
	// The current color modal
	static final public int COLOR_DEFAULT = 1;
	static final public int COLOR_DOMAIN = 2;
	
	// Used by Color Bar 
	Color [] colorBarColors = null;
	int[] colorBarBins = null;
	int colorBarBinMaxSize;
	String[] colorBarStrings = null;
	int[] colorBarStringsY = null;

	int string_space = 0;
	boolean colorBarUsed = false;

	final Color COLORBAR_WORD_COLOR = Color.black;
	final Color COLORBAR_WORD_COLOR_MAP = Color.white;
	final Color DATE_COLOR = Color.black;

	boolean colorKeyUsed = false;

	// Keeps track of the number of files are current being used. 
	int num_files = 1;
	int current_fileIndex =0;

    //For commandline (offscreen mode)
    private boolean offScreenMode=false;
    private int legendWidth=0;

	// ------- Animation Stuff -----------------------------
	// A array of images one for each file.  Which it then flips throw
	// for the animation.
	Image[] slids;
	// The controlling thread
	Thread myThread;
	// The sleep time
	long sleep_length= 1000;
	// the number of times to go throught the slid show;
	int slid_times = 1;	
	// The animator class
	DisplayInterface animator;
	
	// ---------- Values Stuff ------------------------------
	// Used to handle values group stuff 
	ValuesGroup current_group = null;

	final int HISTOGRAM_LENGTH = 5;
	final int LOWER_BOUND_DOUBLE_COLOR = 10;
	final int MAX_SHOW_STRINGS = 10;
	boolean values_set = false;
	Color[] values_colors = null;
	Strings values_strings = null;
	double[] values_doubles = null;
	int[] values_count = null;
	int values_num = 0;
	final String OTHER_STRING = "other";
	
	double spectrum_min =0;
	double spectrum_max = 0;

	// -------- Method Stuff -----------------------------
	// The current Color Method
	static public final int METHOD_NONE = -1;
	static public final int METHOD_DEFAULT = 0;
	int method = METHOD_DEFAULT;
	static public final int METHOD_SPECTRUM = 1;
	static public final int METHOD_MASK = 2;

	// -------- Color Stuff -----------------------------
	ColorMaskFrame colorMaskFrame;

	public Display()
	{
		setBackground(Color.white);
    
		statusBar = new StatusBar(this);
		setMode(ADD_NODE);
		printMessage("");

		dlist = new DList();
		selected = new DList();
		
		check_hover = new Check_Hover(this);
		check_hover.start();

		width = getSize().width;
		height = getSize().height;
		image_changed = true;
	}
    /*
    public Display(Image image, double top_lat_input, double top_lon_input, double bot_lat_input, double bot_lon_input, boolean saveOnly)
    {
	this(image,top_lat_input,top_lon_input,*/

	public Display(Image image, double top_lat_input, double top_lon_input, double bot_lat_input, double bot_lon_input)
	{
		this();
		source_image =  image;
		source_top_lat = top_lat_input;
		source_top_lon = top_lon_input;
		source_bot_lat = bot_lat_input;
		source_bot_lon = bot_lon_input;
	}

	/********************************************************
	* Purpose: To set up the shift created by the scroll bars
	*   Currently not implement because Otter can't seem 
	*   to get the scoll bars to work
	*
	*********************************************************/
	public void Shift_X(float shift) {}
	public void Shift_Y(float shift) {}

	/********************************************************
	* Purpose: To set the current mode and display it in the 
	*  status bar.
	*********************************************************/
	public void setMode(int mode_type)
	{
		if (mode == mode_type)
			return;
		mode = mode_type;

		switch (mode)
		{
			case ADD_NODE: statusBar.setString(ADD_NODE_STR,0);
				break;
			case ADD_LINK: statusBar.setString(ADD_LINK_STR,0);
				break;
			case SELECT: statusBar.setString(SELECT_STR,0);
				break;
			case HIDE_SELECT: statusBar.setString(HIDE_SELECT_STR,0);
				break;
			case ROOT_SELECT: 
				statusBar.setString(ROOT_SELECT_STR,0);
				break;
			case SELECT_TREE: 
				statusBar.setString(SELECT_TREE_STR,0);
				break;
			case PROPERT: statusBar.setString(PROPERT_STR,0);
				break;
			case REMOVE: statusBar.setString(REMOVE_STR,0);
				break;
			case FORMAT: 
				statusBar.setString(FORMAT_STR,0);
				break;
			case MOVE_NAMES: statusBar.setString(MOVE_NAMES_STR,0);
				break;
			case MOVE: statusBar.setString(MOVE_STR,0);
			case ANIMATE: statusBar.setString(ANIMATE_STR,0);
				break;
			case ZOOM: statusBar.setString(ZOOM_STR,0);
				break;
			case STEP_BY_STEP: statusBar.setString(STEP_BY_STEP_STR,0);
				break;
			default:
				mode = ADD_NODE;
				statusBar.setString(ADD_NODE_STR,0);
		}
		repaintStatusBar();
	}

    public void setOffScreenMode(boolean offscreen)
    {
	this.offScreenMode=offscreen;
    }


	/********************************************************
	* Purpose: To get the current mode
	*********************************************************/
	public int getMode() { return mode; }
	
	/********************************************************
	* Purpose: To set the current resolution.
	*********************************************************/
	public void setRes(int res_type)
	{
	    setResNoPaint(res_type);
	    repaintNodes();
	}

	/********************************************************
	* Purpose: To set the current resolution.
	*********************************************************/
	public void setResNoPaint(int res_type)
	{
		res = res_type;
		boolean little_on = false;
		
		if (res == LOW)
		{
			statusBar.setString(LOW_STRING,1);
			little_on = true;
			
		}
		else 
		{
			statusBar.setString(HIGH_STRING,1);
			little_on = false;
		}
		if (dlist == null)
			return;
		DList mylist = dlist.DList();
		mylist.reset();
		while(!mylist.end())
			mylist.next().setLittle(little_on);
		repaintStatusBar();
	}
 	/*******************************************************
	* Purpose: to get the whole current drawing image
	********************************************************/
	public Image getOffscreen()
	{
		return offscreen;
	}

	/********************************************************
	* Purpose: reset the size of the box 
	*********************************************************/
	public void mapSizeChange()
 	{
/*
		if(image == null)
			return;

        	if (worldProducer == null ) {
                	worldProducer = new WorldProducer(source_image,
				top_lon, top_lat, bot_lon, bot_lat, 
				width, height);
			worldProducer.setParent(this);
		}
        	else
                	worldProducer.setSize(width,height);

        	image = createImage(worldProducer);
		image_changed = true;
		size_changed = true;
		xyChange();
*/
 	}

	//I can't figure out why display.size() isn't working
	public int[] displaySize()
	{
		int dsize[] = new int[2];
		dsize[0] = width;
		dsize[1] = height;
		return dsize;
	}
			
	/********************************************************
	* Purpose: Allow for other Classes to print things in the
	*	Status bar.
	*********************************************************/
	public void printMessage(String message)
	{
		statusBar.setString(message,2);
		repaintStatusBar();
	}

	/********************************************************
	* Purpose: Gets the current string in the status Bar
	*********************************************************/
	public String getMessage()
	{
		return(statusBar.getString(2));
	}	
	
	/********************************************************
	* Purpose: Sets the current display list (ie the list of 
	*	nodes and lines)
	*********************************************************/
	public void setList(DList list)
	{
		setListNoPaint(list);
		repaintNodes();
	}

	/********************************************************
	* Purpose:Sets the date for the current list
	*********************************************************/
	public void setDate(Date date_input)
	{
		date = date_input;
	}
	
	/********************************************************
	* Purpose:To draw the date of the file
	*********************************************************/
	public void showDate()
	{
		date_showen = true;
		repaint();
	}
	public void hideDate()
	{
		date_showen = false;
		repaint();
	}
	public void swapDateShow()
	{
		 date_showen = !date_showen;
		repaint();
	}

        public void swapDrawArrows() {
            Arrows.swapDrawArrows();
            repaintNodes();
        }

        public void increaseArrowSize() {
            Arrows.increaseArrowSize();
            repaintNodes();
        }

        public void decreaseArrowSize() {
            Arrows.decreaseArrowSize();
            repaintNodes();
        }

        public void increaseArrowAngle() {
            Arrows.increaseArrowAngle();
            repaintNodes();
        }

        public void decreaseArrowAngle() {
            Arrows.decreaseArrowAngle();
            repaintNodes();
        }

	public void setMap()
	{
                top_lat = source_top_lat;
                top_lon = source_top_lon;
                bot_lat = source_bot_lat;
                bot_lon = source_bot_lon;
                width = getSize().width;
                height = getSize().height;

		// caida To remove the possible of crash caused
		// by lack of image.
/*
		if (source_image != null)
		{
                	worldProducer = new WorldProducer(source_image,
				top_lon, top_lat,bot_lon, bot_lat, 
				width,height);
			worldProducer.setParent(this);
                	image = createImage(worldProducer);
			default_image = image;
			size_changed = false;
			setMap = true;
		}
*/
	}

	public void showAllNames() { showNames(true); }
	public void showRootNames() { showNames(false); }
	protected void showNames(boolean showAll)
	{
                ShiftImage.setShiftImage(true);
		DList mylist = dlist.DList();
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				if (showAll || ((Node)object).isRoot())
					((Node)object).showName();
				else
					((Node)object).hideName();
			}
		}
		repaintNodes();
	}

	public void hideNames()
	{
                ShiftImage.setShiftImage(false);
		DList mylist = dlist.DList();
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
				((Node)object).hideName();
		}
		repaintNodes();
	}

	public void setShowMap(boolean flag)
	{
		showMap = flag;
	}

	public boolean isShowMap()
	{
		return showMap;
	}

	public void setViewGeo(boolean flag)
	{
		viewGeo = flag;
	}
	
	public boolean isViewGeo()
	{
		return viewGeo;
	}

	public boolean isSetMap()
	{
		return setMap;
	}

	public void setMap(boolean flag) {
		setMap = flag;
	}
	
	public void setOnMap()
	{
		DList mylist = dlist.DList();
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
			//object.onMap(isShowMap());
			object.onMap(isViewGeo());
		}
	}

	/********************************************************
	* Purpose: Sets the current display list (ie the list of 
	*	nodes and lines) with out redrawing the image
	*	used when loading.
	*********************************************************/
	public void setListNoPaint(DList list)
	{
		dlist = list;
		selected.clear();
		method = METHOD_DEFAULT;
		values_set = false;
		image_changed = true;
		setMethod(METHOD_DEFAULT);
	}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public DList getList()
	{
		return dlist;
	}

	/********************************************************
	* Purpose: 
	*********************************************************/
	public DList getOrderedList()
	{
		return ordered;
	}

	public void setOrderedList(DList list)
	{
		ordered = list;
	}

	/********************************************************
	* Purpose:To select which fileIndex to use
	*********************************************************/
	public void setFileIndex(int fileIndex)
	{
		setFileIndexNoPaint(fileIndex);

		if (colorBarUsed || colorKeyUsed) 
			colorBySpectrum();
		else
			repaintNodes();
	}
	
	/********************************************************
	* Purpose:To select which fileIndex to use
	*********************************************************/
	public void setFileIndexNoPaint(int fileIndex)
	{
		current_fileIndex = fileIndex;
		DList mylist = dlist.DList();
		mylist.reset();
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			object.setFileIndex(fileIndex);
		}
	}
	/********************************************************
	* Purpose:To set the number of files current in use
	*********************************************************/
	public void setNumFiles(int numFiles)
	{
		num_files = numFiles;
	}
	
	/********************************************************
	* Purpose:To get the number of files the list has
	*********************************************************/
	public int getNumFiles() { return num_files; }

	/********************************************************
	* Purpose:
	*********************************************************/
	public void deselectAll()
	{
		while (!selected.empty())
			selected.pop().setSelect(false);
		while (!nearest_objects.empty())
			nearest_objects.pop().setHighlight(false);
		repaintNodes();
	}

	/********************************************************
	* Purpose:To take in a string find all nodes that contain
	* that string in their name
	*********************************************************/
	public void selectByName(String string)
	{
		printMessage("Searching ...");
		char[] array = string.toCharArray();

		// Remove white space at the front
		int start = 0;
		while (start < array.length-1 && (array[start] == ' ' 
			|| array[start] == '\t'))
			start++;

		// Remove white space at the end and check to see
		// if it is a ip address
		int end = start;
		while (end < array.length && array[end] != ' ' 
			&& array[end] != '\t')
			end++;

		string = new String(array,start,end-start);

		while (!selected.empty())
			selected.pop().setSelect(false);
	
		DList mylist = dlist.DList();
		mylist.reset();
		boolean found = false;
		int num_found = 0;
		while(!mylist.end())
		{
		  DisplayObject object = mylist.next();
		  if (object instanceof Node)
		  {
		    Node node = (Node) object;
		    if (node.getName().indexOf(string) >= 0)
		    {
		       found = true;
		       node.setRoot(true);
		       //node.setSelect(true);
		       //selected.push(node);
		       num_found++;
     		    }
		    else
			node.setRoot(false);	
		  }
		}

		if (!found)
			printMessage("Unable to find any nodes");
		else
			printMessage("Nodes matched:"+ num_found);
		repaintNodes();
	}

	public void lineSize(int size_input) {
		int min = size_input;
		DList mylist = dlist.DList();
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
			if (object instanceof Link) {
				int old = object.getSize();
				if (old + size_input <= -1) {
					if(old -1 < (-1*min))
						min = 1 - old;
				}
			}
                }
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
			if (object instanceof Link || object instanceof Path) {
				int old = object.getSize();
				object.setSize(min+old);
			}
                }
		if (min != size_input)
			printMessage("Link size can't be smaller than 0 ");
		else
			printMessage("");
		repaintNodes();
	}

	public void nodeSize(int size_input) {
		int min = size_input;
		DList mylist = dlist.DList();
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
			if (object instanceof Node) {
				int old = object.getSize();
				if (old + size_input <= 0) {
					if(old -1 < (-1*min))
						min = 1 - old;
				}
			}
                }
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
			if (object instanceof Node) {
				int old = object.getSize();
				object.setSize(min+old);
			}
                }
		if (min != size_input)
			printMessage("Node size can't be smaller than 0 ");
		else
			printMessage("");
		repaintNodes();
	}
	/********************************************************
	* Purpose:To scale all the nodes xy by some scaler
	*********************************************************/
	public void scale(double scaler_input)
	{
		if (isViewGeo()) {
			scaler = 1/scaler_input;
			printMessage("Click the point you are in terested");
		}

		else  {
			printMessage("Zooming ... Please wait");
			zoomNoMap(scaler_input);
			printMessage("");
			recenter();
			repaintNodes();
		}
	}

	public void zoom(double scaler_input)
	{
		double lon = (top_lon + bot_lon)/2;
		double lat = (top_lat + bot_lat)/2;
		zoom(lon,lat,scaler_input);
	}

	public void zoom(double t_lon, double t_lat, double b_lon, double b_lat)
	{
		while(t_lon >= b_lon)
			b_lon += 360;
		double lon = (t_lon + b_lon)/2;
		double lat = (t_lat + b_lat)/2;
		double scaler = (b_lon - t_lon)/(bot_lon - top_lon);
		zoom(lon,lat,scaler);
	}
	
	public void zoom(int x, int y, double scaler_input)
	{
		double latlong[] = FromMap(x,y);
		zoom(latlong[1],latlong[0],scaler_input);
	}

	public void zoom(double m_lon, double m_lat, double scaler_input)
	{
		double lat_size, lon_size;
		if (top_lat < bot_lat)
			lat_size = (bot_lat - top_lat) * scaler_input;	
		else
			lat_size = (top_lat - bot_lat) * scaler_input;
		if (top_lon < bot_lon)
			lon_size = (bot_lon - top_lon) * scaler_input;	
		else
			lon_size = (top_lon - bot_lon) * scaler_input;
		if (lon_size/360 < MAX_ZOOM)
			return;
		if (lat_size >180)
		{
			lon_size = lon_size* 180/lat_size;
			lat_size = 180;
		}
		if (lon_size >360)
		{
			lat_size = lat_size* 360/lon_size;
			lon_size = 360;
		}
		double t_lat = m_lat + lat_size/2;
		double t_lon = m_lon - lon_size/2;
		double b_lat = m_lat - lat_size/2;
		double b_lon = m_lon + lon_size/2;
		
		while (t_lon < -180)
		{
			t_lon += 360;
			b_lon += 360;
		}
		while (t_lon > 180)
		{
			t_lon -= 360;
			b_lon -= 360;
		}
		while (t_lat > 90)
		{
			b_lat -= t_lat -90;
			t_lat = 90;
		}
		while (b_lat < -90)
		{
			t_lat -= b_lat +90;
			b_lat = -90;
		}

		top_lon = t_lon;
		top_lat = t_lat;
		bot_lon = b_lon;
		bot_lat = b_lat;

/*
		if(isShowMap()) {
			printMessage("Loading Image, Please wait....");
			if (worldProducer != null)
				worldProducer.abort();
			worldProducer.setBox(top_lon,top_lat,bot_lon,bot_lat);
                	image = createImage(worldProducer);

			if(size_changed)
				mapSizeChange();
		}
*/
		xyChange();
	}

	public void xyChange()
	{
		DList mylist = dlist.DList();
                mylist.reset();
                while (!mylist.end())
                {
                        DisplayObject object = mylist.next();
                        if (object instanceof Node) {
				double lat = object.getLatitude();
				double lon = object.getLongitude();
				int[] xy = ToMap(lat,lon);
                                ((Node)object).setXY(xy[0],xy[1]);
			}
                }
	}

	public Dimension getImageSize()
	{
		DList mylist = dlist.DList();
		mylist.reset();
		int min_x = 99999;
		int min_y = 99999;
		int max_x = -1;
		int max_y = -1;
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				int x = ((Node)object).getX();
				int y = ((Node)object).getY();
				if (x < min_x)
					min_x = x;
				if (x > max_x)
					max_x = x;
				if (y < min_y)
					min_y = y;
				if (y > max_y)
					max_y = y;
			}
		}
		return new Dimension(max_x - min_x,max_y - min_y);
	}
	public Rectangle getImageExtremeCoords()
	{
		DList mylist = dlist.DList();
		mylist.reset();
		int min_x = 999999;     // Was 50
		int min_y = 999999;     // Was 50
		int max_x = -1;
		int max_y = -1;
                String max_x_node,max_y_node,min_x_node,min_y_node;
                max_x_node = new String("EMPTY");
                max_y_node = new String("EMPTY");
                min_x_node = new String("EMPTY");
                min_y_node = new String("EMPTY");
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				int x = ((Node)object).getX();
				int y = ((Node)object).getY();
				if (x < min_x) {
					min_x = x;
                                        min_x_node = ((Node)object).name;
                                }
				if (x > max_x) {
					max_x = x;
                                        max_x_node = ((Node)object).name;
				} 
                                if (y < min_y) {
					min_y = y;
                                        min_y_node = ((Node)object).name;
                                }
				if (y > max_y) {
					max_y = y;
                                        max_y_node = ((Node)object).name;
                                }
			}
		}
		return new Rectangle(min_x,min_y,max_x - min_x,max_y - min_y);
	}
    /*	public Dimension getKeySize(Dimension dim,
		FontMetrics fm)
	{
		int charHeight = fm.getHeight();
		int charWidth = fm.charWidth(' ');
		int max_height  = dim.height - dim.height;
		int barX = charWidth;
		int stringX = barX + (3*charWidth)/2;
		int barY = charHeight;

		int longest = 0;
		for(int index=0; index < values_colors.length; index++)
		{
			barY += charHeight;
			String string = values_strings.get(index);
			int length = fm.stringWidth(string);
			if (length > longest)
				longest = length;
			barY += charHeight/2;
			if (barY >=max_height)
			{
				barY = charHeight;
				barX +=  longest + 2*charWidth;
				stringX += longest + 2*charWidth;
				longest = 0;
			}
		}
		return (new Dimension(stringX + longest + charWidth,
			dim.height));
	}
    */
 	public void setImageSize(int width_input,int height_input)
	{
		DList mylist = dlist.DList();
		mylist.reset();
		int min_x = 50;
		int min_y = 50;
		int max_x = 0;
		int max_y = 0;
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				int x = ((Node)object).getX();
				int y = ((Node)object).getY();
				if (x < min_x)
					min_x = x;
				if (x > max_x)
					max_x = x;
				if (y < min_y)
					min_y = y;
				if (y > max_y)
					max_y = y;
			}
		}
		if (width_input == max_x - min_x 
				&& height_input == max_y - min_y)
			return;
		printMessage("Zooming ... Please wait");

		double x_scale = width_input/(double)(max_x - min_x);
		double y_scale = height_input/(double)(max_y - min_y);
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				int x = (int)(((Node)object).getX()*x_scale);
				int y = (int)(((Node)object).getY()*y_scale);
				((Node)object).setXY(x,y);
			}
		}

		printMessage("");
		recenter();
		repaintNodes();
	}

	public void zoomNoMap(double scaler_input)
	{
		DList mylist = dlist.DList();
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
				((Node)object).scale(scaler_input);
		}
	}
		
	public int[] ToMap(double lat, double lon)
        {
                Dimension dim = getSize();

                int[] xy = new int[2];
		double boundary = (top_lon + bot_lon)/2 - 180;

		while (lon < boundary ) 
			lon += 360.0;

                xy[0] = (int) (
                        ((lon - top_lon) * width) / (bot_lon - top_lon));
                xy[1] = (int) (
                        ((top_lat - lat) * height) / (top_lat - bot_lat));
                return xy;
        }

	public double[] FromMap(int x, int y)
        {
                Dimension dim = getSize();
                double[] node_lat_lon = new double[2];

                node_lat_lon[0] = top_lat - ((double) y * (top_lat - bot_lat) /
(double) height);
                node_lat_lon[1] = top_lon + ((double) x * (bot_lon - top_lon) /
(double) width);
                return node_lat_lon;
        }

	/********************************************************
	* Purpose:To display all objects step by step 
	*********************************************************/
	public void stepBystepNoPaint(boolean flag)
	{
		stepBystep = flag;	
	}

	public void stepBystep(boolean flag)
	{
		stepBystep = flag;	
		repaintNodes();
	}

	public void stepBystep(int current_step_input)
	{
		stepBystep = true;	
		current_step = current_step_input;
		repaintNodes();
	}

	public boolean isStepByStep()
	{
		return stepBystep;
	}

	/********************************************************
	* Purpose: To set Current Values Group
	*********************************************************/
	public void setGroup(ValuesGroup group) 
	{
		if (group == null)
			return;
		colorKeyUsed = false;
		colorBarUsed = false;
		values_set = false;
		current_group = group;
		DList mylist = dlist.DList();
		mylist.reset();
		while(!mylist.end())
			((DisplayObject)mylist.next()).setGroup(group);
		if (method == METHOD_DEFAULT)
			setMethod(METHOD_SPECTRUM);
		else
		{
			int temp_method = method;
			method = METHOD_DEFAULT;
			setMethod(temp_method);
		}
	}	

	public void setMethod(int method_input)
	{
		if (method == method_input)
		{
			if (method_input == METHOD_MASK)
			{
				colorMaskFrame.toFront();
				colorMaskFrame.show();
			}
			return;
		}
		colorBarUsed = false;
		colorKeyUsed = false;
		method = method_input;
		if (colorMaskFrame != null && method != METHOD_MASK)
			colorMaskFrame.hide();
		switch (method)
		{
			case METHOD_MASK:
				colorByMask();
				break;
			case METHOD_SPECTRUM:
				colorBySpectrum();
				break;
			case METHOD_DEFAULT:
				colorByDefault();
				break;
		}
	}
		

	public ValuesGroup getGroup() { return current_group; } 
			
	/********************************************************
	* Purpose:To color all objects by there Selected Value
	*********************************************************/
	public void colorByMask()
	{
		if (colorMaskFrame == null)
			colorMaskFrame = new ColorMaskFrame(this,dlist);
		colorMaskFrame.toFront();
		colorMaskFrame.show();
		colorMaskFrame.Color(current_group);
	}
			
			
	/********************************************************
	* Purpose:To color all objects by there Selected Value
	*********************************************************/
	public void colorBySpectrum()
	{
		if (current_group == null)
			printMessage("You must select a value to color by");
		else 
		{
			printMessage("Creating Spectrum for "
				+ current_group.getCurrentField()
				+ " .. Please wait");
			if (current_group.getType() == ValuesGroup.DOUBLE)
				colorBySpectrumDouble();
			else if (current_group.getType() == ValuesGroup.STRING)
				colorBySpectrumString();
			if (values_set)
			{
				repaintNodes();
				printMessage("Color by "
					+ current_group.getCurrentField());
			}
		}
	}
	
	/**********************************************************
	* Pupose: To color all objects by their double value
	***********************************************************/
	public void colorBySpectrumDouble()
	{
		DList mylist = dlist.DList();
		if (!values_set)
		{
			spectrum_max = 0;
			spectrum_min = 0;
			if (mylist.top() != null)
				spectrum_max = spectrum_min = 
				   ((DisplayObject)mylist.top()).getDouble();
			if (values_count == null || values_doubles == null)
			{
				int length = mylist.count()/2;
				values_colors = new Color[length];
				values_count = new int[length];
				values_doubles = new double[length];
			}
			values_num = 0;

			mylist.reset();
			boolean first_value = false;
			while(!mylist.end())
			{
			    DisplayObject object = mylist.next();
			    for(int i=0; i < num_files; i++)
			    {
				object.setFileIndex(i);
				if (object.exists() && object.hasValues())
				{
					double value = object.getDouble();

					if (value < spectrum_min)
						spectrum_min = value;
					if (value > spectrum_max)
						spectrum_max = value;

					int index=0;
					while (index < values_num
					  && values_doubles[index] > value)
						index++;
					if (index < values_num &&
					   values_doubles[index] == value)
						values_count[index]++;
					else 
						addValue(index,value);
				}
			    }
			}

			/** SPECIAL HACK ADD HERE *****/
			/** Allows for distict values to be give for doubles
			***	when there a few of them
			****/
			if (values_num < LOWER_BOUND_DOUBLE_COLOR)
			{
				values_strings = new Strings();
				for (int i=0; i < values_num; i++)
				{
				  values_strings.add(
				  String.valueOf((float)values_doubles[i]),
				    values_count[i]);
				}
				colorBySpectrumStringEnd(mylist);
				return;
			}
			
			values_colors = new Color[values_num];
			for (int i=0 ;i < values_num; i++)
			{
				double value = values_doubles[i];
				if (value == Double.NEGATIVE_INFINITY)
					value = Double.MIN_VALUE;
				else if (value == Double.POSITIVE_INFINITY)
					value = Double.MAX_VALUE;
				double normalized = 1-(value-spectrum_min)
				  / (spectrum_max-spectrum_min);
				values_colors[i] = value2Color(normalized);
			}
				
			mylist.reset();
			while(!mylist.end())	
			{
			    DisplayObject object = mylist.next();
			    for(int i=0; i < num_files; i++)
			    {
				object.setFileIndex(i);
				if (object.exists() && object.hasValues())
				{
			 	  double value = object.getDouble();
				  double normalized = ((value-spectrum_min)
				    / (spectrum_max-spectrum_min));
				  object.setValueColor(value2Color(
                                   normalized));
				}
			    }
			    object.setFileIndex(current_fileIndex);
			    object.setToValueColor();
			} 
			colorBarColors = null;
			values_set = true;
		}
		else
		{
			/** SPECIAL HACK ADD HERE *****/
			/** Allows for distict values to be give for doubles
			***	when there a few of them
			****/
			if (values_num < LOWER_BOUND_DOUBLE_COLOR)
			{
				colorBySpectrumStringEnd(mylist);
				return;
			}

			mylist.reset();
			while(!mylist.end())
			{
				DisplayObject object = mylist.next();
				if (object.hasValues())
					object.setToValueColor();
			}
		}
		colorKeyUsed = false;
		colorBarUsed = true;
	}

	/********************************************************
	* Purpose:To color all objects by there top level string
	*********************************************************/
	public void colorBySpectrumString()
	{
		DList mylist = dlist.DList();
		if (!values_set)
		{
			values_strings = new Strings();
			mylist.reset();
			while(!mylist.end())
			{
			    DisplayObject object = mylist.next();
			    if (object.hasValues())
			    {
				for(int i=0; i < num_files; i++)
				{
					String string = object.getString();
					values_strings.addUnique(string);
				}
			    	if (false && values_strings.length() > 200)
			    	{
			   		 printMessage("Too many values to "
						+"display a Spectrum");
					colorBarUsed = false;
			    		colorKeyUsed = false;
					return;
			    	}
			    }
			    else
				object.setToDefaultColor();
			}
			if (values_strings.length() <= MAX_SHOW_STRINGS)
			{
				values_strings.sort();
			}
			else
			{
			  String[] strings = new String[MAX_SHOW_STRINGS];
			  int[] strings_num = new int[MAX_SHOW_STRINGS];
			  int total = 0;
			  
			  int current_num_strings = 0;
			  for(int index=0; index < values_strings.length(); 
			    index++)
			  {
			    int dst = current_num_strings;
			    int num = values_strings.getOccurances(index);
			    String string = values_strings.get(index);
			    total += num;
			    while (0 < dst && num > strings_num[dst-1])
			      dst--;
			    if (current_num_strings < MAX_SHOW_STRINGS)
				current_num_strings++;
			    if (dst < current_num_strings)
			    {
                              for (int i=current_num_strings-1; i > dst; i--)
			      {
                                strings[i] = strings[i-1];
			        strings_num[i] = strings_num[i-1];
                              }
			      strings[dst] = string;
	  		      strings_num[dst] = num;
                            }
			  }
			  
			  values_strings = new Strings();
			  for (int index=0; index < MAX_SHOW_STRINGS-1; index++)
			  {
                            values_strings.add(strings[index],
			      strings_num[index]);
			    total -= strings_num[index];
			  }
			  values_strings.add(OTHER_STRING,total);
			}
		}
		colorBySpectrumStringEnd(mylist);
	}
		
	public void colorBySpectrumStringEnd(DList mylist)
	{
		if (!values_set)
		{
			int length = values_strings.length();
			values_colors = createSpectrum(length);
			
			/*
			int unknown_index = values_strings.addUnique("Unknown");

			length++;
			Color[] temp = new Color[length];
			for (int i=0; i < unknown_index; i++)
				temp[i] = values_colors[i];
			temp[unknown_index] = Link.DEFAULT_COLOR;
			for (int i=unknown_index+1; i < length; i++)
				temp[i] = values_colors[i];
			values_colors = temp;
			*/
			mylist.reset();
			int other_index = values_strings.indexOf(OTHER_STRING);
			while(!mylist.end())
			{
			    DisplayObject object = mylist.next();	
			    if (object.hasValues())
			    {
				for(int i=0; i < num_files; i++)
				{
				     object.setFileIndex(i);
				     if (object.hasValues())
 				     {
					String string ="";
					if (current_group.getType() 
					  == ValuesGroup.DOUBLE)
					  string = String.valueOf(
						(float)object.getDouble());
					else if (current_group.getType() 
					  == ValuesGroup.STRING)
					  string = object.getString();
					int index = values_strings.indexOf(string);
					if (index > -1)
					  object.setValueColor(
						values_colors[index]);
				        else
					  object.setValueColor(
						values_colors[other_index]);
				    }
				}
			   	object.setFileIndex(current_fileIndex);
			   	object.setToValueColor();
			   }
			   else
				object.setToDefaultColor();
		  	}
			values_set = true;
		}
		else
		{
			mylist.reset();
			while(!mylist.end())
			{
				DisplayObject object = mylist.next();
				if (object.hasValues())
					object.setToValueColor();
				else
					object.setToDefaultColor();
			}
		}
		colorBarUsed = false;
		colorKeyUsed = true;
	}

	protected void addValue(int index, double double_input)
	{
		if (values_count.length <= values_num)
		{
			int length = (int)(values_num*1.2);
			if (length == values_count.length)
				length = values_count.length*10;
			int[] temp_c = values_count;
			double[] temp_d = values_doubles;
			
			
			values_count = new int[length];
			values_doubles = new double[length];

			for(int i=0; i< temp_c.length; i++)
			{
				values_count[i] = temp_c[i];
				values_doubles[i] = temp_d[i];
			}
		}

		for (int i=values_num;i > index;i--)
		{
			values_count[i] = values_count[i-1];
			values_doubles[i] = values_doubles[i-1];
		}
		values_count[index] = 1;
		values_doubles[index] = double_input;
		values_num++;
	}
			

	/*******************************************************
	* Purpose: To color by defualt colors
	*******************************************************/
	public void colorByDefault()
	{
		DList mylist = dlist.DList();
		mylist.reset();
		while(!mylist.end())
			mylist.next().setToDefaultColor();
		colorBarUsed = false;
		colorKeyUsed = false;
		printMessage("Colored by Default");
		repaintNodes();
	}

	protected static Color value2Color(double input)
        {
          double hue, sat,bri;
 
          if (input <= 0.000001)
            input = .000001;
          double temp = Math.log( (120*input)+1)/Math.log(121);
//	  double temp = input;
	  hue = (4+5*temp)/8;
          //sat = (2 + 8*temp)/10;
	  sat = 1;
          bri = 1;
          return new Color( Color.HSBtoRGB(
              (float) hue,
              (float) sat,
              (float) bri)
              );

	}

	protected static Color[] createSpectrum(int length)
	{
		Color[] colors = new Color[length];

		if ( length <= 13 )
		{ 

			// These are draw in reverse order
			// You can always spend the time to find out 
			// why later, right ?
			Color[] key_colors = {
				new Color(0x3399ff),
				new Color(0x009900),

				new Color(0x00ff00),
				new Color(0xffcc00),

				new Color(0xffcccc),
				new Color(0xff6600),

				new Color(0xff0000),
				new Color(0x6666fc),

				new Color(0xff66ff),
				new Color(0xccccff),

				new Color(0x999966),
				new Color(0x66ccff),

				new Color(0xffff00),
				};
			for (int i=0; i < length;i++)
				colors[length-1-i] = key_colors[i];
			return colors;
		}
		for (int i=0; i < length; i++)
		{
			float hue = (float)(.9*(i))
				/(float)length;
			//float sat = (float)(70-((hue*100)%30))/(float)70;
			//float bright = 10 - (float) (.2*i)/length;
			colors[i] = Color.getHSBColor(hue,1,1);
		}
		
		return  colors;
	}
		
	/*********************************************************
	* Purpose:To paint out the current date of the network
	**********************************************************/
	public String dateString()
	{
		String string = "no date";
		if (date != null)
		{
			switch (date.getMonth())
			{
				case 0: string = "JAN"; break;
				case 1: string = "FEB"; break;
				case 2: string = "MAR"; break;
				case 3: string = "APR"; break;
				case 4: string = "MAY"; break;
				case 5: string = "JUN"; break;
				case 6: string = "JUL"; break;
				case 7: string = "AGU"; break;
				case 8: string = "SEP"; break;
				case 9: string = "OCT"; break;
				case 10: string = "NOV"; break;
				case 11: string = "DEC"; break;
			}
			string = string + " " + date.getDate() + ", ";
			string = string + (date.getYear() + 1900);
		}
		return string;
	}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public void Clear()
	{
		dlist.clear();
		colorByDefault();
		printMessage("Network Cleared");
	}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public synchronized void paint(Graphics g)
	{
		Dimension dim = getSize();
		if ((width != dim.width || height != dim.height)) {
			size_changed = true;
			width = dim.width;
			height = dim.height;
			if(isShowMap())
				mapSizeChange();
		}
		FontMetrics fm = getFontMetrics(getFont());
		statusBar.draw(g);

                if (ready) {
        	    paintNodes(g, dim);
                }
	}


	/********************************************************
	* Purpose:
	*********************************************************/
	public void repaintStatusBar()
	{
		Font font = getFont();
		// This checks to see if the font has been initialized
		if ((font != null)&&(!offScreenMode))
		{

			Graphics g = getGraphics();
			statusBar.draw(g);

/*
			Dimension dim = getSize();
			FontMetrics fm = getFontMetrics(font);
			int old_height = statusBar.getHeight();
			statusBar.draw(g,dim,fm);
			int new_height = statusBar.getHeight();
			if (old_height > new_height)
			{
				g.clipRect(0,dim.height-old_height,dim.width,
					old_height-new_height);
				paint(g);
			}
*/
		}
	}

	/********************************************************
	* Purpose:
	*********************************************************/
    public void repaintNodes() 
    {
	if (offScreenMode)
	    return;
	image_changed = true;
	if (getGraphics()==null) System.out.println("Jeepers! getGraphics() returned null in Display.repaintNodes()");
	if (getSize()==null) System.out.println("Jeepers! getSize() returned null in Display.repaintNodes()");

	paintNodes(getGraphics(), getSize());
    }
	/********************************************************
	* prash-mess:
	*********************************************************/
	public void repaintNodes(Graphics g) 
	{ 
		image_changed = true;
		paintNodes(g, getSize()); 
	}
	/********************************************************
	* Purpose:
	*********************************************************/
	protected void paintNodes(Graphics g, Dimension dim)
	{
		// Redraw the offscreen image (ie that complete image)
		// if you have to
	    if (g==null) System.out.println("Jeepers! Display.paintNodes() called with null graphics object");
		if (image_changed == true && !(isShowMap() || isViewGeo())) 
			CheckOffScreenSize();
		if (isShowMap() || isViewGeo()) {
			x_shift = 0;
			y_shift = 0;
		}
		if (node_width < dim.width)
			node_width = dim.width;
		if (node_height < dim.height)
			node_height = dim.height;
		if (offscreen == null 
			|| node_width > offscreen_width
			|| node_height > offscreen_height)
		{
			image_changed = true;
			if (node_width > offscreen_width)
				offscreen_width = node_width;
			if (node_height > offscreen_height)
				offscreen_height = node_height;
			if (offscreen != null)
				offscreen.flush();
			offscreen = createImage(offscreen_width
				,offscreen_height);
		}
		if (image_changed)
			redrawOffscreen();

		int height = statusBar.getHeight();
		int clip_height = dim.height - height;

		// Redraw from the original image onto the clipped image
		g.clipRect(0,0,dim.width,clip_height);
		g.setColor(getBackground());
		g.drawImage(offscreen,x_shift,y_shift,this);

		//if(this.isStepByStep())
		//	drawStepByStep(g,x_shift,y_shift);

		// The following four ifs take care of the times when the image 
		// of the network is shifted off the screen. leaving blank space.
		// Not to fill background with bizarre color 
		g.setColor(getBackground());
		if (x_shift > 0)
			g.fillRect(0,0,x_shift,clip_height);
		if (dim.width > offscreen_width + x_shift)
		{
			int erase_x = offscreen_width + x_shift;
			int erase_width = dim.width - erase_x;
			g.fillRect(erase_x,0,erase_width,clip_height);
		}
		if (y_shift > 0)
			g.fillRect(0,0,dim.width,y_shift);
		if (clip_height > offscreen_height + y_shift)
		{
			int erase_y = offscreen_height + y_shift;
			int erase_height = clip_height - erase_y;
			g.fillRect(0,erase_y,dim.width,erase_height);
		}
		if (height <= 0)
		{
			g.clipRect(0,0,dim.width,dim.height);
			FontMetrics fm = getFontMetrics(getFont());
			statusBar.draw(g);
		}

		FontMetrics fm = getFontMetrics(getFont());
		paintColor(g,dim,fm);
		
		if (date_showen)
			paintDate(g,dim,fm);
	}

	public void paintColor(Graphics g, Dimension dim, FontMetrics fm)
	{
		if (colorBarUsed)
			paintColorBar(g,dim,fm);
		if (colorKeyUsed)
			paintColorKey(g,dim,fm);
	}

	Node[] nodes;
	protected void drawStepByStep(Graphics g, int x_shift, int y_shift)
	{
		DList mylist = ordered.DList();
		int rank = -1;
		int color_index = 0;
		int count = mylist.count(); 
		//when reverse order is needed
		//int index = count; 
		int index = 0; 
		if (nodes == null || nodes.length != count)
			nodes = new Node[count];
		DList links = new DList();
		DList node_list = new DList();

		mylist.reset();
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			if(object instanceof Node)
			{
				Node node = (Node)object;
				nodes[index++] = node;
			//when reverse order is needed
			//	nodes[--index] = node;
			}
		}

		for(int i=0; i< count; i ++)
			nodes[i].changeXY(x_shift,y_shift);

		this.setOnMap();
		int node_rank = 0;
		for(int i = 0; i < count; i ++)
		{
			node_rank = nodes[i].getRank();
			if(node_rank > current_step) {
				if(current_step > 0)
					printMessage ("Current level: " +current_step);
				else if (current_step == 0)
					printMessage ("Click [Next] to start");

				break;
			}

                        Dimension dim = getSize();

			if(rank != node_rank && i > 0 )
			{
				//try { Thread.sleep(2000);
				//} catch (InterruptedException e){
				//System.err.println(e.toString());}

				color_index++;
				links.reset();
				while(!links.empty())
					(links.pop()).draw(g, dim, 
                                                    getLatLon(),paintCounter);
				node_list.reset();
				while(!node_list.end())
					(node_list.next()).draw(g, dim,
                                                    getLatLon(),paintCounter);
			}
			else
			{
				DList node_link = nodes[i].getLink();
				node_link.reset();
				while(!node_link.end())
					links.enqueue(node_link.next());
			}
			nodes[i].draw(g, dim ,getLatLon());
			node_list.enqueue(nodes[i]);
			rank = node_rank;
		}

		if (node_rank <= current_step)
				printMessage ("The end of hierarchy");

		//stepBystepNoPaint(false);
		image_changed = true;

		for(int i=0; i< count; i ++)
			nodes[i].changeXY(x_shift,y_shift);
	}
	protected synchronized void redrawOffscreen()
	{
		printMessage("Drawing Nodes .. Please wait");
		Dimension dim = getSize();

		Graphics g = offscreen.getGraphics();
		FontMetrics fm = getFontMetrics(getFont());

		// Need to clear the slate clear
		g.setColor(getBackground());
		g.fillRect(0,0,offscreen_width,offscreen_height);

		if (image != null && isShowMap())
			g.drawImage(image,0,0,this);

		DList mylist; 
        
		this.setOnMap();
		if(!this.isStepByStep())
		{
			mylist = dlist.DList();
			mylist.reset();
			while(!mylist.end())
			{
				DisplayObject object = mylist.next();

                                if (object instanceof Link) {
                                    object.draw(g, dim, getLatLon(), 
                                                paintCounter);
                                } else if(object instanceof Node) {
                                    object.draw(g, dim, getLatLon(), 
                                                paintCounter);
                                } else {
    				    object.draw(g, dim ,getLatLon());
                                }
			}
		}
		else
			drawStepByStep(g,0,0);

                paintCounter++;
		image_changed = false;
		printMessage("");
	}
	
	
	/********************************************************
	* Purpose: This draws the nodes/lines onto a graphic
	*	passed it, but it normalizes them to be just
	*	off the edge of the image.
	********************************************************/
	public void drawPostscript(PsGenerator postscript,
                                   boolean withKey, boolean hasWindow)
	{
		if (dlist.empty())
			return;

                DisplayObject object = null;
                int pageX = postscript.getWidth();
                int pageY = postscript.getHeight();
                float x_shift = 0;
                float y_shift = 0;
                float x_scale = 1; 
                float y_scale = 1;
                int legendW = legendWidth;

                int xBorder = postscript.getXBorder();
                int yBorder = postscript.getYBorder();

		printMessage("Drawing Postscript .. Please wait");

                Rectangle rect = getImageExtremeCoords();
		DList mylist; 

                Dimension dim;
                dim = getImageSize();

                Dimension pageSize = new Dimension(pageX, pageY);

		FontMetrics fm = postscript.getFontMetrics(
    			                            postscript.getFont());

                postscript.setShift(0,20);
                if (withKey) {
        	    paintColor(postscript,dim,fm);
                } else if (!hasWindow) {
                    paintColor(postscript,dim,fm);
                } else {
                    legendW = 0;
                }
                postscript.setShift(0,0);

                int effectivePageX;
                int effectivePageY;

                if (ShiftImage.getShiftImage()) {
    		    effectivePageX=pageX-380;            // 80 for as graph (?)
                    effectivePageY=pageY-110;
                } else {
                    effectivePageX=pageX-80;            // 80 for as graph (?)
                    effectivePageY=pageY-60;
                }

		x_scale=(float)effectivePageX/((float)(dim.width));
		y_scale=(float)effectivePageY/((float)(dim.height));

		//we want the smaller one
		if (x_scale<y_scale)
		    y_scale=x_scale;
		else
		    x_scale=y_scale;

		//the shift values are POST scaling
		x_shift= -(rect.x*x_scale);
		y_shift= -(rect.y*y_scale);

                if (ShiftImage.getShiftImage()) {
                    postscript.setShift(185, 100);
                } else {
                    postscript.setShift(50, 50);
                }
		
		mylist = dlist.DList();
		mylist.reset();

		while (!mylist.end())
		{
		    object = mylist.next();
		    object.draw(postscript,x_shift,y_shift,x_scale,y_scale,
				dim, getLatLon(), true);
		}
		postscript.setShift(0,0);
		
		printMessage("");
	}

	/***********************************************************
	* Purpose: To draw a color key along the left side of the screen
	************************************************************/
	protected void paintColorKey(Graphics g, Dimension dim,
		FontMetrics fm)
	{
	    /** This should probably be here...
	    if (g instanceof PsGenerator){
		System.err.println("Setting don't scale in PsGenerator");
		((PsGenerator)g).setShouldScale(false);
	    }*/
	    
	    int charHeight = fm.getHeight();
	    int charWidth = fm.charWidth(' ');
	    int max_height  = dim.height - 2*statusBar.getHeight();
	    int barX = charWidth;
	    int stringX = barX + (3*charWidth)/2;
	    int barY = charHeight;
	    
	    if(isShowMap())
		g.setColor(COLORBAR_WORD_COLOR_MAP);
	    else
		g.setColor(COLORBAR_WORD_COLOR);
	    if (current_group != null) {
		g.drawString(current_group.getTitle()+": "
			     + current_group.getCurrentField(),barX, barY);

		if (barX+fm.stringWidth(current_group.getTitle()+": "
				   + current_group.getCurrentField()) > legendWidth)
		    legendWidth=barX+fm.stringWidth(current_group.getTitle()+": "
					       + current_group.getCurrentField());
	    }

	    // Returns if there are too many objects
	    if (values_colors.length > MAX_SHOW_STRINGS)
		return;

	    barY += charHeight/2;
	    int longest = 0;
	    
	    int max_occurances = 0;
	    for(int index=values_colors.length-1; index >=0 ; index--)
	    {
		    int occurances = values_strings.getOccurances(index);
		    if (max_occurances < occurances)
			max_occurances = occurances;
	    }
	    
	    int last_loop_index=values_colors.length-1;
	    for(int index=values_colors.length-1; index >=0 ; index--)
	    {
		g.setColor(values_colors[index]);
		g.fillRect(barX,barY,charWidth,charHeight);
		barY += charHeight;
		g.setColor(COLORBAR_WORD_COLOR);

		if(isShowMap())
		    g.setColor(COLORBAR_WORD_COLOR_MAP);
		else
		    g.setColor(COLORBAR_WORD_COLOR);

		String string = values_strings.get(index);

		int length = fm.stringWidth(string);

		if (length > longest)
		    longest = length;
		if (length> legendWidth)
		    legendWidth=length;

		g.drawString (string,stringX,barY);

		if ((stringX+fm.stringWidth(string))>legendWidth)
		    legendWidth=stringX+fm.stringWidth(string);

		barY += charHeight/2;


// BELOW FOLLOWS COMPLETLY BOGUS CODE that screws up the placing/drawing
// of the histogram by an elf who shall remain
// Nameless. I'll leave it in (commented) in case something is usefull.
// -- Wilfred.

/*		if (barY >=max_height || index == 0)
		{
		    barY = charHeight;
		    barX +=  longest + 2*charWidth;
		    stringX += longest + 2*charWidth;
				//******************************************
				// This draws the Histogram
				//******************************************
		    for (int i=last_loop_index; i >= index; i--)
		    {
			int occurances = values_strings.getOccurances(i);
			int bar_length = 
			    (HISTOGRAM_LENGTH*charWidth
			     *occurances)
			    / max_occurances;
			g.setColor(values_colors[i]);
			g.fillRect(barX, barY, bar_length, 
				   charHeight);

			if (bar_length+barX>legendWidth)
			    legendWidth=bar_length+barX;

			barY += charHeight;

			g.setColor(COLORBAR_WORD_COLOR);
			g.drawString(String.valueOf(occurances),
				     barX, barY);

			if (barX+fm.stringWidth(String.valueOf(occurances))>legendWidth) {
			    legendWidth=barX+fm.stringWidth(String.valueOf(occurances));
			    
			    barY += charHeight/2;
				//barY += charHeight + charHeight/2;
			}
			barY = charHeight;
			barX +=  6*charWidth;
			stringX += 6*charWidth;
			longest = 0;
			last_loop_index = index-1;
		    }
		}

// These two make the bars horizontal instead of vertical....
// (That's why we don't need it...)
//		barY = charHeight;
//		barX += longest +2*charWidth;
*/


	    /**
	    if (g instanceof PsGenerator){
		System.err.println("Setting do scale in PsGenerator");
		((PsGenerator)g).setShouldScale(true);
	    }   */
	    }
	}

    
	/***********************************************************
	* Purpose: To draw a colorbar along the left side of the screen
	************************************************************/
	protected void paintColorBar(Graphics g, Dimension dim, 
		FontMetrics fm)
	{
		int charWidth = fm.charWidth(' ');
		int charHeight = fm.getHeight();
		int rec_width = 2*charWidth;
		int shift_x = 0;
		int shift_y = this.height;
		int longest = 0;
  		int bar_top = dim.height/8;
		int bar_height = (3*dim.height)/4;
		int bar_bottom = bar_top + bar_height;

		if (colorBarColors == null 
			|| colorBarColors.length != bar_height)
		{

			colorBarColors = new Color[bar_height];
			for (int i=0; i < colorBarColors.length;i++)
			{
				double normalized = 
  				  ((double)1.0)-((double)(colorBarColors.length-i)/ 
						 colorBarColors.length);
				colorBarColors[i] = value2Color(normalized);
			}
			
			colorBarBins = new int[bar_height];
			colorBarBinMaxSize = 0;
			DList mylist = dlist.DList();
			mylist.reset();
			while (!mylist.end())
			{
				DisplayObject object = mylist.next();
				if (object.hasValues())
				{
					double value = object.getDouble();
 					int i = (int) ( 
				 		((colorBarBins.length-1)
						*(value - spectrum_min))
			         		/(spectrum_max) );

					if (i >= colorBarBins.length)
						i = colorBarBins.length-1;
					colorBarBins[i]++;
					if (colorBarBins[i] >colorBarBinMaxSize)
						colorBarBinMaxSize 
							= colorBarBins[i];
				}
			}
			if (colorBarBins[colorBarBins.length-1] == 0)
			  colorBarBins[colorBarBins.length-1] = 1;
				
			int power = 
			  (int)(Math.log(bar_height/(2*charHeight))/Math.log(2));
			int spaceing = bar_height/(1<<power);
			int num_strings =(bar_height/spaceing)+1;
			colorBarStrings = new String[num_strings];
			colorBarStringsY = new int[num_strings];
			for (int i=0;i < colorBarStrings.length;i++)
			{
			  double temp = (double)i/(colorBarStrings.length-1);
			  colorBarStrings[i] = String.valueOf( (int)(temp
			    *(spectrum_max-spectrum_min) + spectrum_min) );
			  colorBarStringsY[i] = (int)(temp*bar_height + 
				bar_top);
			}
		}

		int barX = charWidth;
		int barX2 = 2*barX + barX;
     		int barY = bar_top - charHeight/2;
		int index=0;

                if (barX2>legendWidth)
		        legendWidth=barX2;


		g.setColor(COLORBAR_WORD_COLOR);
		if(isShowMap())
			g.setColor(COLORBAR_WORD_COLOR_MAP);
		else
			g.setColor(COLORBAR_WORD_COLOR);
		if (current_group != null){	
			g.drawString(current_group.getTitle() +": " 
				+ current_group.getCurrentField(), barX,barY);
			if ((fm.stringWidth(current_group.getTitle()+": "+current_group.getCurrentField())+barX)>legendWidth)
			    legendWidth=(fm.stringWidth(current_group.getTitle()+": "+current_group.getCurrentField())+barX);
		}

		//---------------------
		// Draw the key
		//---------------------
		barY += charHeight/2;
		while (index < colorBarColors.length)
		{
			g.setColor(colorBarColors[index]);
			//			g.drawLine(barX,barY,barX2,barY);
			g.drawRect(barX,barY,barX2-barX,1);
			barY++;
			index++;
		}
		
		//---------------------
		// Draw the labels on the key
		//---------------------
		barX2 += barX/2;
		int max_length = 0;
		for (int i=0; i < colorBarStrings.length;i++)
		{
			g.setColor(COLORBAR_WORD_COLOR);
			if(isShowMap())
				g.setColor(COLORBAR_WORD_COLOR_MAP);
			else
				g.setColor(COLORBAR_WORD_COLOR);

			barY = colorBarStringsY[i];
			g.drawLine(barX,barY,barX2,barY);
			barY += charHeight/2;

			g.setColor(COLORBAR_WORD_COLOR);
			g.drawString (colorBarStrings[i],barX2, barY);
			int len = fm.stringWidth(colorBarStrings[i]);
			if (len > max_length)
			    max_length = len;
			if (len>legendWidth)
			    legendWidth=len;
		}
		
		barX = (int)(1.5*charWidth) + max_length + barX2;
		barY = bar_top;

		//*******************************************************
		// This draws the histograph
		//*******************************************************
		for (index=0; index < colorBarBins.length; index++)
		{
			g.setColor(colorBarColors[index]);
			if (colorBarBins[index] > 0)
			  g.drawLine(barX, barY, barX + 
			     (HISTOGRAM_LENGTH*charWidth*colorBarBins[index])/
			     colorBarBinMaxSize, barY);
			//so that we don't overlap the image with the key
			if ((barX+(HISTOGRAM_LENGTH*charWidth*
				  colorBarBins[colorBarBins.length-1])
			     /colorBarBinMaxSize)
			    >this.legendWidth){
			    this.legendWidth=barX+(HISTOGRAM_LENGTH*charWidth*
				colorBarBins[colorBarBins.length-1])/colorBarBinMaxSize;
			}
			barY++;

		}
		
	}
		
    

	/***********************************************************
	* Purpose: Prints the date on the screen
	************************************************************/
	protected void paintDate(Graphics g, Dimension dim, FontMetrics fm)
	{
		String string = dateString();
		int length = fm.stringWidth(string) + 2*fm.stringWidth(" ");
		int height = fm.getHeight();

		g.setColor(DATE_COLOR);
		g.drawString(string,dim.width-length,(int)(1.5*height));
	}

	protected void printDate()
	{
		printMessage("Date:" + dateString());
	}
	
	/***********************************************************
	* Purpose: to find out if any of the nodes will be draw outside
	* 	of the offscreen image.  If so it finds where the 
	*	new bondray should be.
	************************************************************/
	protected void CheckOffScreenSize()
	{
		DList mylist = dlist.DList();
		mylist.reset();
		int min_x = 99999;
		int min_y = 99999;
		int max_x = -1;
		int max_y = -1;
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				int x = ((Node)object).getX();
				int y = ((Node)object).getY();
				if (x < min_x)
					min_x = x;
				if (x > max_x)
					max_x = x;
				if (y < min_y)
					min_y = y;
				if (y > max_y)
					max_y = y;
			}
		}

		if (min_x != 99999 || min_y != 99999)
		{
			mylist.reset();
                        min_x = -1 * min_x;
                        min_y = -1 * min_y;

			while(!mylist.end())
			{
				DisplayObject object = mylist.next();
				if (object instanceof Node)
				   ((Node)object).changeXY(min_x,min_y);
			}
			max_x += min_x;
			max_y += min_y;
		}

                if (max_x > node_width-ShiftImage.getXShift() ||
                    max_y > node_height-ShiftImage.getYShift()) {
			node_width = max_x + (2*ShiftImage.getXShift());
			node_height = max_y + (2*ShiftImage.getYShift());
		}
		//if (max_x - min_x > width)
	//		zoomNoMap(width/(max_x - min_x));
	//	if (max_x - min_x > 0 && (max_y - min_y)*width/(max_x - min_x) > height)
	//		zoomNoMap(height/(max_y - min_y));
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void recenter()
	{
		// It doesn't make sense when nodes are formatted with geo 
		if (isShowMap() || isViewGeo())
			return;
		CheckOffScreenSize();
                boolean set = false;
                int xy[] = null;
                int num_nodes = 0;
                int center_x = 0;
                int center_y = 0;
		DList mylist = dlist.DList();
		mylist.reset();
                while(!mylist.end())
                {
                        DisplayObject object = mylist.next();
                        if (object instanceof Node)
                        {
                                num_nodes++;
                                if (set)
                                {
                                        center_x += ((Node)object).getX();
                                        center_y += ((Node)object).getY();
                                }
                                else
                                {
                                        center_x = ((Node)object).getX();
                                        center_y = ((Node)object).getY();
                                        set = true;
                                }
                        }
                }
		if (num_nodes > 0)
		{ 
			Dimension dim = getSize();
			x_shift = dim.width/2 - (center_x/num_nodes);
			y_shift = dim.height/2 - (center_y/num_nodes);
		}
	}

	/********************************************************
	* Purpose:To handle the mouse events
	*********************************************************/
	Date last_move_date = new Date();
	public boolean handleEvent(Event event)
	{
		// Shifts the give x,y by the shift
        	event.x -= (x_shift+ShiftImage.getXShift());
        	event.y -= (y_shift+ShiftImage.getYShift());

		if (mode == FORMAT || mode == ANIMATE)
			return super.handleEvent(event);
		switch(event.id)
		{
			case Event.ACTION_EVENT:
				if (event.target == colorMaskFrame)
					repaintNodes();
				break;
			case Event.MOUSE_MOVE:
				if (mode != ZOOM && mode != STEP_BY_STEP 
					&& mode != MOVE) {
					check_hover.set(event.x,event.y);
				}
				break;
			case Event.MOUSE_DOWN:
				old_x = event.x;
				old_y = event.y;
				if (mode == ADD_NODE)
					AddNode(event.x, event.y);
				else if (mode == ADD_LINK)
					AddLink(event.x,event.y);
				else if (mode == SELECT || mode == SELECT_TREE
					|| mode == MOVE_NAMES)
					DownSelect(event.x,event.y);
				else if (mode == PROPERT)
					Propert(event.x,event.y);
				else if (mode == REMOVE)
					Remove(event.x,event.y);
				else if (mode == ROOT_SELECT)
					RootSelect(event.x,event.y);
				else if (mode == ZOOM) {
					zoom(event.x,event.y,scaler);
					repaintNodes();
                			printMessage("");
					setMode(SELECT);
				}
				break;
			case Event.MOUSE_DRAG:
			{
				mouse_dragged = true;
				if (mode == ADD_LINK)
					AddLink_Drag(event.x,event.y);
				else if (((mode == SELECT 
					|| mode == SELECT_TREE)
					&& mouse_down_selected != null)
					|| mode == ADD_NODE)
					ChangeXY(event.x, event.y);
				else if (mode == MOVE_NAMES)
					MoveNames(event.x,event.y);
				else if (mode  != ZOOM) 
					Move(event.x,event.y);
				break;
			}
			case Event.MOUSE_UP:
				if (mode == ADD_LINK)
					AddLink_Up(event.x,event.y);
				else if (mode == SELECT || mode == SELECT_TREE)
					UpSelect(event.x,event.y);
				break;
		}	
		return super.handleEvent(event);
	}

	/**************************************************
	* Purpose: To display the objects information
	*
	* nearest_object: is used to reduce needless redrawing
	* of the information
	***************************************************/
	DList nearest_objects = new DList();
	public void GetString(int x, int y)
	{
		DisplayObject object = near(x,y);
		if (object != null)
			printMessage(object.getInfo());
/*
		DList objects = nearObjects(x,y);
		String answer = new String(); 
		boolean equal = true;
		objects.reset();
		while (!objects.end() && equal)
		{
			if(!nearest_objects.exists(objects.next()))
				equal = false;
		}
		if (equal)
			return;

		nearest_objects.reset();
 		while (!nearest_objects.end())
                {
                        DisplayObject old = nearest_objects.next();
                        if (old.isHighlighted()) {
				//old.setHighlight(false);
                                nearest_objects.drop(old);
			}
		}

		if (objects.empty())
			printMessage(null);
		else {
			objects.reset();
			while(!objects.end()) {
				DisplayObject object = objects.next();
				if (answer.length() == 0)
					answer = object.getInfo();
				else
					answer = answer + " : " + object.getInfo();
				//object.setHighlight(true);
			}
			printMessage(answer);	
		}
		nearest_objects = objects;
		//repaintNodes();
*/
	}
	/********************************************************
	* Purpose:
	*********************************************************/
	int counter = 0;
	public void AddNode(int x, int y)
	{
		DisplayObject object = near(x,y);
		// Don't allow nodes to be added on top of each other
		// If some does attempt to do so. Toggle the node
		// to Selected = !Selected
		if (object instanceof Node)
		{
			if (object.isSelected())
			{
				object.setSelect(false);
				selected.drop(object);
			}
			else
			{
				object.setSelect(true);
				selected.add(object);
			}
			repaintNodes();
			return;
		}
		Node node = new Node("Node "+ counter++,x,y);
		node.setSelect(true);
		dlist.add(node);
		selected.add(node);
		printMessage(node.getInfo());	

		repaintNodes();
	}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public void AddLink(int x, int y) 
	{
		DisplayObject object = near(x,y);
		if (object instanceof Node)
		{
			addlink_node = new Node("AddLink Node",x,y);
			addlink_link = new Link((Node)object,addlink_node
				,DisplayObject.DIR_ONE);
			dlist.add(addlink_link);
		}
		else
		{
			addlink_node = null;
			addlink_link = null;
		}
		image_changed = true;
		repaintNodes();
	}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public void AddLink_Drag(int x, int y)
	{
		if (addlink_node != null && addlink_link != null)
			addlink_node.setXY(x,y);
		image_changed = true;
		repaintNodes();
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void AddLink_Up(int x, int y)
	{
		DisplayObject object = near(x,y);
		if (addlink_link != null && object instanceof Node
			&& addlink_link.getParent() != object)
			addlink_link.setChild((Node)object);
		else
			dlist.drop(addlink_link);
		addlink_link = null;
		addlink_node = null;
		image_changed = true;
		repaintNodes();
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void DownSelect(int x, int y)
	{
		mouse_down_selected = near(x,y);
		mouse_dragged = false;
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void UpSelect(int x, int y)
	{
		if (mouse_dragged)
			return;
		int near_distance = -1;
		int distance;
		DisplayObject object = mouse_down_selected;
		if (object != null && !(object instanceof Link))
		{
			if (mode == SELECT)
			{
				if (object.isSelected())
				{
					object.setSelect(false);
					selected.drop(object);
				} 
				else
				{
					object.setSelect(true);
					selected.enqueue(object);
					if (!(object instanceof Node))
					{
					 dlist.drop(object);
					 DList mylist = dlist.DList();
					 mylist.reset();
					 while (!(mylist.where() 
						instanceof Node))
					  mylist.next();
				 	 mylist.insert(object);
					}
				}
			}
			else
			  ((Node)object).selectTree(!object.isSelected()
				,selected);
			// Makes sure the the current state of the object 
			// is correct
			image_changed = true;
			repaintNodes();
		}
		if (object != null)
			printMessage(object.getInfo());
	}
	/********************************************************
	* Purpose:
	*********************************************************/
	public void Propert(int x, int y) {}
	
	/********************************************************
	* Purpose:
	*********************************************************/
	public void Remove(int x,int y)
	{
		DisplayObject object = near(x,y);
		if (object != null)
		{
			selected.drop(object);
			dlist.drop(object);
			DisplayObject[] objects = object.remove();
			for (int index=0; index < objects.length; index++)
			{
				object = objects[index];
				object.remove();
				dlist.drop(object);
				selected.drop(object); 
			}
			image_changed = true;
			repaintNodes();
		}
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void RootSelect(int x, int y)
	{
		DisplayObject object = near(x,y);
		if (object instanceof Node)
		{
			((Node)object).swapRoot();
			repaintNodes();
		}
	}	

	/********************************************************
	* Purpose:
	*********************************************************/
	public void ChangeXY(int x, int y)
	{
		int delta_x = x - old_x;
                old_x = x;
                int delta_y = y - old_y;
                old_y = y;
		if (mouse_down_selected != null 
			&& !mouse_down_selected.isSelected())
		{
			image_changed = true;
			mouse_down_selected.changeXY(delta_x,delta_y);
		}
		else
		{
			selected.reset();
			while (!selected.end())
			{
				DisplayObject object = selected.next();
				if (!object.isSelected())
					selected.drop(object);
				else
				{
					image_changed = true;
					object.changeXY(delta_x, delta_y);
				}
			}
		}
		if (image_changed)
			repaintNodes();
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void MoveNames(int x, int y)
	{
		int delta_x = x - old_x;
                old_x = x;
                int delta_y = y - old_y;
                //old_y = y;
		if (mouse_down_selected instanceof Node)
		{
			if (delta_y> 0)
				((Node)mouse_down_selected).nameClockWise();
			else
				((Node)mouse_down_selected).nameCounterClock();
	/*
			Node node = (Node) mouse_down_selected;
			int delta_x = x - node.getX();
			int delta_y = node.getY() - y;
			double base_angle;
                        if (delta_x == 0 && delta_y == 0)
                                base_angle = 1.5*Math.PI;
                        else if (delta_x == 0)
                                base_angle = 0;
                        else if (delta_y == 0)
                                if (delta_x > 0)
                                        base_angle = .5*Math.PI;
                                else
                                        base_angle = 1.5*Math.PI;
                        else
                                base_angle = Math.atan ((delta_x)
                                        /(double)(delta_y));
			base_angle += .5*Math.PI;
                        if (delta_y < 0)
                                base_angle += Math.PI;

			node.setNameLocation(base_angle);
	*/
			repaintNodes();
		}
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	public void Move(int x, int y)
	{
		int delta_x = x -old_x;
		int delta_y = y - old_y;
		if(isViewGeo()) {
			double[] old_latlon = FromMap(old_x,old_y);
			double[] latlon = FromMap(x,y);
			double delta_lon = latlon[1] - old_latlon[1];
			double delta_lat = latlon[0] - old_latlon[0];
			zoom (top_lon-delta_lon,top_lat-delta_lat,bot_lon-delta_lon,bot_lat-delta_lat);
			repaintNodes();
                        printMessage("");
		}
		else {
		
			x_shift += delta_x;
			y_shift += delta_y;
			repaint();
		}

		old_x = x - delta_x;
		old_y = y - delta_y;
	}

	/********************************************************
	* Purpose:
	*********************************************************/
	protected DisplayObject near(int x, int y)
 	{
		double near_distance = -1;
		double distance;
		DisplayObject near_object = null;
		DList mylist = dlist.DList();
		mylist.reset();
		while(!mylist.end())
		{
			DisplayObject object = mylist.next();
			distance = object.near(x,y);
			if (near_distance < 0 && distance > 0)
			{
				near_distance = distance;
				near_object = object;
			}
			else if (distance > 0 && (
			(!(near_object instanceof Node)
			&& object instanceof Node)
			|| distance < near_distance))
			{
				near_distance = distance;
				near_object = object;
			}
		}
		return near_object;
	}

	protected DList nearObjects(int x, int y)
 	{
		double near_distance = -1;
		double distance;
		DList near_objects = new DList();
		DisplayObject near_object = null;
		dlist.reset();
		while(!dlist.end())
		{
			DisplayObject object = dlist.next();
			distance = object.near(x,y);
			if (near_distance < 0 && distance > 0)
			{
				near_distance = distance;
				near_objects.enqueue(object);
				near_object = object;
			}
			else if (distance > 0 && (
			(!(near_object instanceof Node)
			&& object instanceof Node)
			|| distance <= near_distance))
			{
				near_distance = distance;
				near_objects.enqueue(object);
				near_object = object;
			}
		}
		near_objects.reset();
		while(!near_objects.end())
		{
			DisplayObject object = near_objects.next();
			distance = object.near(x,y);
			if (distance > near_distance || (near_object instanceof
			    	Node && object instanceof Link))
				near_objects.drop(object);
		}
		return near_objects;
	}
	/********************************************************
	* Purpose:To allow for the Display to sleep
	*********************************************************/
	public void setAnimator(DisplayInterface animator_input)
	{
		animator = animator_input;
	}
	
	/********************************************************
	* Purpose:To allow for the Display to sleep
	*********************************************************/
	public void setThread(Thread thread)
	{
		myThread = thread;
	}

	/********************************************************
	* Purpose:Sets the amout of time that it would sleep.
	*********************************************************/
	public void setSleep(long sleep)
	{
		sleep_length = sleep;
	}
	
	/********************************************************
	* Purpose:Sets the number of times it should loop
	*********************************************************/
	public void setTimes(int time)
	{
		slid_times = time;
	}

	/********************************************************
	* Purpose: To get  boundary  lat/longs -jjung
	*********************************************************/
	public double[] getLatLon()
	{
		double[] lat_lon = new double[4];
		lat_lon[0] = top_lat;
		lat_lon[1] = top_lon;
		lat_lon[2] = bot_lat; 
		lat_lon[3] = bot_lon; 
		return lat_lon;
	}
	public void setLatLon(double t_lat, double t_lon, double b_lat, 
			      double b_lon)
	{
		source_top_lat = top_lat = t_lat;
		source_top_lon = top_lon = t_lon;
		source_bot_lat = bot_lat = b_lat;
		source_bot_lon = bot_lon = b_lon;
	}
	
	
	/********************************************************
	* Purpose:To create a group of images that will be presented
	* in a side show.
	*********************************************************/
	public void run()
	{
		int current_mode = mode;
		setMode(ANIMATE);
		slids = new Image[num_files];
		current_fileIndex = 0;
		int times = 0;
		while(times < slid_times)
		{
			animator.setFileIndex(current_fileIndex);
			if (slids[current_fileIndex] == null)
			{
				offscreen = null;
				setFileIndex(current_fileIndex);
				slids[current_fileIndex] = offscreen;
			}
			else
			{
				offscreen = slids[current_fileIndex];
				repaint();
			}
			try {
				myThread.sleep(sleep_length);
			} catch (InterruptedException exc) {
				System.err.println ("Display.run():"
					+ exc.toString());
			}
			current_fileIndex = (current_fileIndex+1)
				%num_files;
			if (current_fileIndex == 0)
				times++;
		}
		
		setMode(current_mode);
	}

	public void ipPull()
	{
		DList mylist = dlist.DList();
		mylist.reset();
		final long mask_0 = 256*256*256*256;
		final long mask_3 = mask_0 - 256;
		final long mask_2 = mask_0 - 256*256;
		Node[] nodes = new Node[mylist.count()];
		int num_nodes = 0;
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
				//&& !((Node)object).isRoot())
				nodes[num_nodes++] = (Node) object;
		}
		for (int i=0; i < 2;i++)
		for (int x=0; x < num_nodes; x++)
  		{
		  Node node = nodes[x];
		  int node_size = node.getSize();
		  for (int y=x; y < num_nodes;y++)
		  {
			Node before = nodes[y];
			long ip_node = node.getIP();
			long ip_before = before.getIP();
			double pull = 0;
			if ((ip_node & mask_3) == (ip_before & mask_3))
				pull = 2;
			//else if ((ip_node & mask_2) == (ip_before & mask_2))
				//pull = 1;
			if (pull > 0)
			{
				int size = (node_size + before.getSize())/4;
				int X = before.getX()- node.getX();
				int Y = before.getY()- node.getY();
				double Z = Math.sqrt(X*X+Y*Y);
				if (Z > size*2)
				{
					int slop_x = (int)((pull*size*X)/Z);
					int slop_y = (int)((pull*size*Y)/Z);
					node.changeXY(slop_x,slop_y);
					before.changeXY(-slop_x,-slop_y);
				}
			}
				
		  }
                }
	}
}

class Check_Hover extends Thread
{
	protected final int WAIT = 10;
	protected boolean set = false;
	protected int x,y;
	protected Display display;
	public Check_Hover(Display display_input)
	{
                super("Check_HoverThread");
		display = display_input;
	}
		
	public void run()
	{
	   try 
	   {
		while(true)
		{
			if (set) {
				display.GetString(x,y);
			}
			set = false;
			sleep(WAIT);
		}
	   } catch (InterruptedException e) {
		System.err.println("Hover_check.run():" + e.toString());
	   }
	}

	synchronized public void set(int x_input,int y_input)
	{
		x = x_input;
		y = y_input;
		set = true;
	}
}
