package caida.otter;
import java.awt.*;
import java.lang.Math;
import java.util.Random;

/**************************************************************************
* File: Format.java	Class: Format
* Goal: To arrange the nodes in a more meaningful fashion.
* 	It arranges the nodes of the passed Display using a depth 
*	first algorithm.
*
* Written: Bradley Huffaker             (12/20/97)
*
* Modified: Jaeyeon Jung		(2/13/98)
* To arrange nodes in a World Map 
*
* For:Cooperative Association for Internet Data Analysis
***************************************************************************
***************************************************************************
By accessing this software, FORMAT, you are duly informed of and 
agree to be bound by the conditions described below in this notice:
 
This software product, FORMAT, is developed by Bradley Huffaker, and
copyrighted(C) 1998 by the University of California, San Diego (UCSD),
with all rights reserved.  UCSD administers the NSF grant to CAIDA,
number NCR-9711092, under which this code was developed. 
 
There is no charge for FORMAT software. You can redistribute it and/or
modify it under the terms of the GNU General Public License, v. 2 dated
June 1991 which is incorporated by reference herein.  FORMAT is
distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that the use of
it will not infringe on any third party's intellectual property
rights.
 
You should have received a copy of the GNU GPL along with the FORMAT
program.  Copies can also be obtained from
http://www.gnu.org/copyleft/gpl.html or by writing to
 
                        University of California, San Diego
                        SDSC/CAIDA
                        9500 Gilman Dr., MS-0505
                        La Jolla, CA 92093 - 0505  USA
 
  Or contact INFO@CAIDA.ORG
**************************************************************************/    
public class Format extends Thread
{
	// The list of DisplayObjects to format
	Display display;
	int oldMode;
		
	// The size of the current node
	int size;
	// Used for random effect
	Random random = new Random();

	// The differt ways to format
	static final int FORCE = 0;
	static final int ROOT = 1;
	static final int GEO = 2;
	static final int SITE1 = 3;
	static final int FIND_ROOTS = 4;
	static final int IP = 5;
	static final int FIXED = 6;
	static final int VALUE = 7;

	protected int mode = FORCE;

	static final double MAX_GROWTH = 10;
	static final int SIZE = 6;
	static final double PUSH = (SIZE*5)*(SIZE*5)*(SIZE*5)*(SIZE*5);
	static final double PULL = SIZE*SIZE;

	// The default size of the Root nodes
	final int ROOT_SIZE = 8;

	// This is used between Animator and Format
	// If it is null it is ingnored
	FormatInterface parent;

	// This is the amount of size to leave on the width to allow
	// for the color key
	static final int EXPECTED_KEY_WIDTH = 100;

	public Format(Display display_input)
	{
                super("FormatDisplayThread");

		display = display_input;

		DList dlist = display.getList();
		dlist.reset();
		while (!dlist.end()) {
			DisplayObject object = dlist.next();
			if(object instanceof Node)
				((Node)object).unSetXY();
		}
		while (!dlist.end()) {
			DisplayObject object = dlist.next();
				object.setLatLon(false);
		}
	}

	public void setMode(int input) { mode = input; }
	public int getMode() {return mode; }

	public void setParent(FormatInterface input)
	{
		parent = input;
	}

	public void printMessage(String msg)
	{
		if (parent == null)
			display.printMessage(msg);
		else
			parent.printMessage(msg);
	}
	
	public void run()
	{
		oldMode = display.getMode();
		display.setMode(Display.FORMAT);
		printMessage("Formatting");
		switch (mode)
		{
			case FIXED:
				ProcessFixed();
				break;
			case FORCE:
				ProcessPullPush();
				break;
			case FIND_ROOTS:
				FindRoots();
			case ROOT:
			{
				DList queue = SetUpRoots();
				ProcessRoot(queue);
				break;
			}
			case SITE1:
			{
				DList queue = SetUpGeos();
				ProcessRoot(queue);
				break;
			}
			case IP:
			{
				SetUpIP();
				break;
			}
			case VALUE:
			{
				/*
				ValuesGroup group = display.getGroup();
				if (group != null 
				  && group.getType() != ValuesGroup.DOUBLE)
				{
				    display.setMode(oldMode);
				    printMessage("ERROR:Must be double values");
				    return;
				}
				*/
				SetupValueBased();
				break;
			}
			//jjung
			//jjung
			case GEO:
				ProcessGeo();
				break;
			default:
				System.err.println("Unknown type of format");
				break;
		}
		printMessage("Finished Formatting");
		MakePositive();
		display.recenter();
		if (parent != null)
			parent.finished();
		else 
			display.colorByDefault();
		display.setMode(oldMode);
	}
/*********************** Process Fixed ***********************************/

	protected void ProcessFixed()
	{
		DList nodes = new DList();
		boolean boundaries_set = false;
		double x_min = 0;
		double x_max = 0;
		double y_min = 0;
		double y_max = 0;

		DList mylist = display.getList().DList();
		Dimension dim = display.size();
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
			{
				double x = ((Node)object).getX();
				double y = ((Node)object).getY();
				if (boundaries_set)
				{
					if (x < x_min)
						x_min = x;
					if (x > x_max)
						x_max = x;
					if (y < y_min)
						y_min = y;
					if (y > y_max)
						y_max = y;
				}
				else
				{
					x_min = x;
					x_max = x;
					y_min = y;	
					y_max = y;
					boundaries_set = true;
				}
				
				object.show();
				nodes.push(object);
			}
		}
		
		/*
		if (nodes.count() < 1 || (x_min == x_max 
			&& y_max == y_min))
			return SetUpRoots();
		*/
		if (x_min == x_max)
			x_min = x_max -1;
		if (y_min == y_max)
			y_min = y_max -1;

		int width = dim.width-20 - EXPECTED_KEY_WIDTH;
		int height = dim.height-20;
		int x_base = width;
		int y_base = height;
		double diff_x = x_max - x_min;
		double diff_y = y_max - y_min;
		double x_scalar = ((double)width)/diff_x;
		double y_scalar = ((double)height)/diff_y;

		if (x_scalar > y_scalar)
		{
			x_scalar = y_scalar;
			x_base = (int)((width)-(diff_x*x_scalar))/2;
		}
		else
		{
			y_scalar = x_scalar;
			y_base = (int)((height)-(diff_y*y_scalar))/2;
		}

		y_scalar *= -1;

		int avg_x = 0;
		int avg_y = 0;
		nodes.reset();
		while(!nodes.end())
		{
			Node node = (Node) nodes.next();
			double node_x = node.getX();
			double node_y = node.getY();
			int x = (int)((node_x-x_min)*x_scalar) + x_base;
			int y = (int)((node_y-y_min)*y_scalar) + y_base;
			avg_x += x;
			avg_y += y;
			node.setXY(x,y);
		}
	}

/*********************** Find Roots ***********************************/

	protected void FindRoots()
	{
		DList dlist = display.getList().DList();
		boolean one_found = true;
		DList nodes  = new DList();
		DList roots = new DList();
		dlist.reset();
		// Divide the list into nodes and roots
		while (!dlist.end())
		{
			DisplayObject object = dlist.next();
			if (object instanceof Node)
			{
				if (((Node)object).isRoot())
					roots.push(object);
				else
				{
					nodes.push(object);
					((Node)object).unSetXY();
				}
			}
		}
		
		// Find All the nodes which can be reached from the current 
		// roots
		roots.reset();
		while (!roots.end())
		{
			ReachNodes((Node)roots.next());
		}

		// Final all nodes which have not been reached and then make
		//  some of them roots till all nodes reached
		while (one_found)
		{
			one_found = false;
			nodes.reset();
			Node node = null;
			Node found = null;
			int largest_degree = 0;
			while (!nodes.end())
			{
				node = (Node) nodes.next();
				int degree = node.getNumNeighbors();
				if (!node.isSetXY() && degree > largest_degree)
				{
					one_found = true;
					largest_degree = degree;
					found = node;
				}
			}
			if (one_found)
			{
				found.setRoot(true);
				ReachNodes(found);
				nodes.drop(found);
			}
					
/*
			while (!nodes.end() && !one_found)
			{
				node = (Node) nodes.next();
				if (!node.isSetXY())
					one_found = true;
			}
			if (one_found)
			{
				node.setRoot(true);
				ReachNodes(node);
				nodes.drop(node);
			}
*/
		}
	}
	
	protected void ReachNodes(Node node)
	{
		if (node.isSetXY())
			return;
		node.setXY();
		Node[] neighbors = node.getNeighbors();
		for(int i=0; i < neighbors.length;i++)
			if (!neighbors[i].isSetXY())
				ReachNodes(neighbors[i]);
	}
	
/*********************** Root Methods ***********************************/
	
	// The basic limiting angle
	final double ANGLE_SCOPE = Math.PI;

	protected void ProcessRoot(DList queue)
	{
		Dimension dim = display.size();
		int center_x = dim.width/2;
		int center_y = dim.height/2;
		//jae
		DList ordered = new DList();
	
		if (queue == null)
			return;
		queue.reset();
		while(!queue.end()) {
			DisplayObject object = queue.next();
			if(((Node)object).getRank()==1)
				ordered.enqueue(object);
		}	
		while (!queue.empty())
		{
			double angle_scope = ANGLE_SCOPE;
			Node top = (Node) queue.pop();
			int x = top.getX();
			int y = top.getY();
			int rank = top.getRank();
				
			double delta_x = x-center_x;
			double delta_y = y-center_y;
			double base_angle;
			if (delta_x == 0 && delta_y == 0)
			{
				base_angle = 1.5*Math.PI;
				angle_scope = 2*Math.PI;
			}	
			else if (delta_x == 0)
				base_angle = 0;
			else if (delta_y == 0)
				if (delta_x > 0)
			 		base_angle = .5*Math.PI;	
				else
			 		base_angle = 1.5*Math.PI;	
			else
				base_angle = Math.atan ((delta_x)
					/(double)(delta_y));
			if (delta_y < 0)
				base_angle += Math.PI;
/*			
System.err.println(top.getString() + "  x:"+ x + "-"+center_x +"="+ delta_x +
"  y:" +y + "-"+center_y +"="+ delta_y );
 System.err.println(" Angle:" + (base_angle/Math.PI));
*/

			int size = top.getSize();
			Node[] neighbors = top.getNeighbors();
			int current = 0;
			int num_neighbors = neighbors.length;

			while (current< num_neighbors)
			{
				int next = current+1;
				String name = neighbors[current].getName();
				boolean set = neighbors[current].isSetXY();
				while (next < num_neighbors)
				{
				  String next_name = neighbors[next].getName();
				  boolean next_set = neighbors[next].isSetXY();
				  // This is to sort them and move
				  // all set nodes to the bottom
				  if (set || (!next_set
				    && name.compareTo(next_name) > 0))
				  {
					Node temp = neighbors[next];
					neighbors[next] = neighbors[current];
					neighbors[current] = temp;
					name = next_name;
					set = next_set;
				  }
				  next++;
				}
				current++;
			}
			// This should remove any nodes which have already
			// had their xy set
			
			// This removes any node from the end of the
			// list which are already set.  this is so 
			// the last node is unset.
			while (num_neighbors > 0 
				&& neighbors[num_neighbors-1].isSetXY()) 
			{
				num_neighbors--;
			}
/*
current=0;
System.err.println("	"+top.getName()+" " + neighbors.length);
while (current < num_neighbors)
{
	System.err.println( neighbors[current].getName() + " " 
		+ neighbors[current].isSetXY() );
	current++;
}
System.err.println("----");
/*
			// This removes the those neighbors that 
			// have already be set.
			int where = 0;
			while (where < num_neighbors)
			{
				if (neighbors[where].isSetXY())
				{
				  num_neighbors--;
				  neighbors[where] = neighbors[num_neighbors];
				}
				else
					where++;
			}	
*/

/*
			if (num_neighbors%2 == 0)
				base_angle = base_angle 
				  + (angle_scope/2*num_neighbors);
*/
			if (num_neighbors %2 == 0)
				base_angle += Math.PI/(4*num_neighbors);
			// The array of children for the current node
			// this array will be given to top at the end 
			// of the for loop
			Node[] children = new Node[num_neighbors];
				
			// Here we set all of the nodes neighbors
			// which have not yet set and added them
			// to the bottom of the queue.  so that 
			// the new nodes neighbors can be set.
			for (int index=0;index< num_neighbors; index++)
			{
			    Node node = neighbors[index];
			    children[index] = node;
			    if (!node.isSetXY())
			    {
			      double length=size*2 + 
				2*node.getNumNeighbors()+num_neighbors;
			      double angle = 0;
				
			      // This will have the angle swap be between 
			      // the potive and negative sides of the base
			      // angle
			      if (index%2 == 0)
				angle = base_angle 
				  + (angle_scope*(index/2)/num_neighbors);
			      else
				angle = base_angle 
				  - (angle_scope*((index+1)/2)/num_neighbors);

			      node.setXY(x+(int)(length*Math.sin(angle)),
			 		y+(int)(length*Math.cos(angle)));
			      node.setNameLocation(angle);
			      node.setRank(rank+1);
			 	
/*
if (index == 0 )
{
 node.setColor(Color.green);
 
 System.err.println(top.getString() +"->"+node.getString() + " length:" 
  + length);
 System.err.println(" x:"+ x + "-"+center_x +"="+ delta_x +
	"  y:" +y + "-"+center_y +"="+ delta_y );
 System.err.println(" Angle:" + (angle/Math.PI) +"\n"
  + " Base angle:" + (base_angle/Math.PI) +"\n"
  + " x:"+ x+"+"+ (int)(length*Math.sin(angle))
  + " y:" +y+"+"+ (int)(length*Math.cos(angle)));
}
*/

				node.setSize(node.getSize());
			/*
			      if (size > 3)
				node.setSize(size-1);
			      else 
			        node.setSize(2);
			*/
			      queue.enqueue(node);

			      ordered.enqueue((DisplayObject)node);
			    }
			}
			top.setChildren(children);
		}
		display.setOrderedList(ordered);
	}

	protected DList SetUpRoots()
	{
		Dimension dim = display.size();
	
		DList queue = new DList();

		DList dlist = display.getList();
		DList roots = new DList();

		// This is a special case if there are not roots.
		Node most= null;
		int most_size = 0;
		int most_root = 0;
		int most_count = 0;

		dlist.reset();
		while (!dlist.end())
		{
		    DisplayObject object = dlist.next();
		    if (object instanceof Node)
  		    {
			Node node = (Node) object;
			int count = node.getNumNeighbors();
			node.unSetXY();

			if (node.isRoot())
			{
				if (count > most_root)
					most_root = count;
				int size = node.getSize();
				if (size > most_size)
					most_size = size;
				roots.push(node);
			}
			if (count > most_count)
			{
				most = node;
				most_count = count;
			}
		    }
		}
		if (roots.count() <= 0)
			if (most == null)
				return null;
			else
				roots.push(most);	
			
		if (roots.count() == 1)
		{
			Node root = (Node) roots.pop();
			root.setToDefault();
			root.setXY(dim.width/2,dim.height/2);
			root.setNameLocation(1.5*Math.PI);
			root.setRank(1);
			queue.enqueue(root);
		}
		else
		{
			int count = roots.count();
			int length = (int)(
				(count*most_size)+most_count/2);
			int shortest = dim.width/3;
			if (dim.height/3 < shortest)
				shortest = dim.height/3;
			if (length > shortest)
				length = shortest;
			for (int index=1; index <= count; index++)
			{
				Node root = (Node) roots.pop();
				double angle = (index*2*Math.PI)/count;
			 	int my_len = length 
					+ root.getNumNeighbors();
				int width = (int)(my_len* Math.cos(angle));
				int height = (int)(my_len* Math.sin(angle));
				root.setToDefault();
				root.setXY(dim.width/2+ width, dim.height/2+height);
				root.setNameLocation(angle);
				root.setRank(1);
				queue.enqueue(root);
			}
		}
		return queue;
	}

/*********************** Geo Method 1 **********************************/

	protected DList SetUpGeos()
	{
		DList queue = new DList();
		DList sites = new DList();
		boolean boundaries_set = false;
		double lat_min = 0;
		double lat_max = 0;
		double lon_min = 0;
		double lon_max = 0;

		DList mylist = display.getList().DList();
		Dimension dim = display.size();
System.out.println("Format.SetUpGeos.dim:" + dim);
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node
				&& ((Node)object).isRoot()
				&& ((Node)object).isGeoSet())
			{
				object.setToDefault();
				double lat = object.getLatitude();
				double lon = object.getLongitude();
				if (boundaries_set)
				{
					if (lat < lat_min)
						lat_min = lat;
					if (lat > lat_max)
						lat_max = lat;
					if (lon < lon_min)
						lon_min = lon;
					if (lon > lon_max)
						lon_max = lon;
				}
				else
				{
					lat_min = lat;
					lat_max = lat;
					lon_min = lon;	
					lon_max = lon;
					boundaries_set = true;
				}
				
				object.show();
				sites.push(object);
			}
			else if (object instanceof Node)
				((Node)object).unSetXY();
		}
		
		if (sites.count() < 1 || (lat_min == lat_max 
			&& lon_max == lon_min))
			return SetUpRoots();
		if (lat_min == lat_max)
			lat_min = lat_max -1;
		if (lon_min == lon_max)
			lon_min = lon_max -1;

		int x_base = dim.width;
		int y_base = dim.height;
		/*
		int x_base = dim.width/2;
		int y_base = dim.height/2;
		/*	When dimension where distoried to make things fit better
		double x_scalar = ((double)dim.width/2)/(lon_max-lon_min);
		double y_scalar = ((double)dim.height/2)/(-lat_max+lat_min);
		*/
		double diff_lon = lon_max - lon_min;
		double diff_lat = lat_max - lat_min;
		double x_scalar = ((double)dim.width/2)/diff_lon;
		double y_scalar = ((double)dim.height/2)/diff_lat;
		if (x_scalar > y_scalar)
		{
			x_scalar = y_scalar;
			x_base = (int)(dim.width-(diff_lon*x_scalar))/2;
		}
		else
		{
			y_scalar = x_scalar;
			y_base = (int)(dim.height-(diff_lat*y_scalar))/2;
		}
		y_scalar *= -1;
		sites.reset();
		Node[] sites_array = new Node[sites.count()];
		int index = 0;

		int avg_x = 0;
		int avg_y = 0;
		while(!sites.empty())
		{
			Node site = (Node) sites.pop();
			sites_array[index++] = site;
			double lat = site.getLatitude();
			double lon = site.getLongitude();
			int x = (int)((lon-lon_min)*x_scalar) + x_base;
			int y = (int)((lat-lat_min)*y_scalar) + y_base;
			avg_x += x;
			avg_y += y;
			site.setXY(x,y);
			queue.push(site);
/*
			Node[] nodes = site.getNodes();
			for(int i=0; i < nodes.length;i++)
				queue.push(nodes[i]);
*/
		}

/*
		if (sites_array.length > 0)
		{
			int distance = (int)(sites_array[0].getSize()*.5);
			if (distance < 5)
				distance = 5;
			PushApart(distance,sites_array); 
		}
*/

		// This moves them back to the center so that the trees
		// grows correctly
		int shift_x = dim.width/2 - avg_x/sites_array.length;
		int shift_y = dim.height/2 - avg_y/sites_array.length;
		for (int i=0; i < sites_array.length;i++)
			sites_array[i].changeXY(shift_x,shift_y);
		
		return queue;
	}

	protected void PushApart(int size, Node[] nodes)
	{
		boolean found_one = true;
		int push_differnt = 0;
		while (found_one)
		{
		  found_one = false;
		  for (int i=0;i < nodes.length;i++)
		    for (int j=(i+1);j < nodes.length;j++)
		    {
			int x1 = nodes[i].getX();
			int x2 = nodes[j].getX();
			int x_x = x1-x2;
			int y1 = nodes[i].getY();
			int y2 = nodes[j].getY();
			int y_y = y1-y2;
			if (x_x == 0 && y_y == 0)
			{
				switch (push_differnt)
				{
					case 0:
						y_y = 1;
						break;
					case 1:
						x_x = 1;
						y_y = 1;
						break;
					case 2:
						x_x = 1;
						break;
					case 3:
						x_x = 1;
						y_y = -1;
						break;
					case 4:
						y_y = -1;
						break;
					case 5:
						x_x = -1;
						y_y = -1;
						break;
					case 6:
						x_x = -1;
						break;
					case 7:
						x_x = -1;
						y_y = 1;
						break;
				}
				push_differnt = (push_differnt+1)%8;
			}
			double length = (Math.sqrt((x_x)*(x_x)
				+(y_y)*(y_y)))/2;
			double shift = size - length;
			if (shift > 1)
			{
				nodes[i].setColor(Color.blue);
				nodes[j].setColor(Color.blue);
				int x_shift = (int)((shift*x_x)/length);
				int y_shift = (int)((shift*y_y)/length);
				nodes[i].setXY(x1+x_shift,y1+y_shift);
				nodes[j].setXY(x2-x_shift,y2-y_shift);
				found_one = true;
			}
		     }
		}
	}
	/********************  Make all values Postive***************/

	protected void MakePositive()
	{
		DList dlist = display.getList().DList();
		dlist.reset();
		int min_x = 0;
		int min_y = 0;
		
		DList nodes = new DList();
		while (!dlist.end())
		{
			DisplayObject object = dlist.next();
			if (object instanceof Node)
			{
				Node node = (Node) object;
				int x = node.getX();
				int y = node.getY();
				if (x < min_x)
					min_x = x;
				if (y < min_y)
					min_y = y;
				nodes.push(node);
			}
		}
		int shift_x = 30 - min_x;
		int shift_y = 30 - min_y;
		if (shift_x != 0 || shift_y != 0)
		{
			nodes.reset();
			while (!nodes.end())
				((Node)nodes.next()).changeXY(shift_x,shift_y);
		}
	}

	/***********************  IP Cluster Method ********************/

	protected void SetUpIP()
	{
		Dimension dim = display.size();
		DList dlist = display.getList().DList();
		DList nodes = new DList();
		dlist.reset();
		while (!dlist.end())
		{
			DisplayObject object = dlist.next();
			if (object instanceof Node)
			{
				((Node)object).setToDefault();
				//((Node)object).setRoot(false);
				nodes.push(object);
			}
		}
		ProcessIP(2,dim.width/2,dim.height/2,nodes);
	}

	protected void ProcessIP(int depth, int center_x,
		int center_y, DList nodes)
	{
		if (nodes.count() == 0)
			return;
		int length = nodes.count()/2+2;
		if (depth == 0)// || length < 4)
		{
			int count = nodes.count();
			int radius = nodes.count()/2 
				+ ((Node)nodes.top()).getSize();
			int current_node = 0;
			nodes.reset();
			while (!nodes.end())
			{
				Node node  = (Node) nodes.next();
				double angle = (2*Math.PI*current_node)
					/count + Math.PI/2;
				int x = center_x +(int)
						(radius*Math.sin(angle));
				int y = center_y + (int)
						(radius*Math.cos(angle));
				node.setNameLocation(angle);
				node.setXY(x,y);
				node.setRank(0);
				current_node++;
			}
			return;
		}
			
		long mask = 256;
		for (int i=1; i < depth; i++)
			mask *= 256;
		mask = 256*256*256*256 + 1 - mask;
		int num_lists=0;
		DList[] lists = new DList[length];
		long[] masks = new long[length];
		nodes.reset();
		while(!nodes.end())
		{
			Node node = (Node) nodes.next();
			int index=0;
			long ip = node.getIP() & mask;
			while ( index < lists.length &&
				lists[index] != null && masks[index] != ip)
				index++;
			if (index == length)
			{
				length = (int)(length*1.5);
				DList[] temp_l = lists;
				long[] temp_m = masks;
				lists = new DList[length];
				masks = new long[length];
				for(int i=0;i < temp_l.length;i++)
				{
					lists[i] = temp_l[i];
					masks[i] = temp_m[i];
				}
			}
			if (lists[index] == null)
			{
				lists[index] = new DList();
				lists[index].push(node);
				masks[index] = ip;
				num_lists++;
			}
			else
				lists[index].push(node);
		}
		for (int i=0;i < num_lists;i++)
		{
			Node node  = (Node) nodes.next();
			double angle = (2*Math.PI*i)
				/num_lists + Math.PI/2;
			length = lists[i].count() +num_lists;
			int x = center_x + (int)(length*Math.sin(angle));
			int y = center_y + (int)(length*Math.cos(angle));
			ProcessIP(depth-1,x,y,lists[i]);
		}
	}
				
/*********************** Value Based  ***********************************/

	protected void SetupValueBased()
	{
		Dimension dim = display.size();
		DList dlist = display.getList().DList();
		
		DList nodes_unset = new DList();
		
		int num_root_neighbors = -1;

		Node root = null;
		Node node = null;
		int num_nodes = 0;
		dlist.reset();
		while (!dlist.end())
		{
			DisplayObject object = dlist.next();
			if (object instanceof Node)
			{
				num_nodes++;
				node = (Node) object;
				nodes_unset.push(node);
				if (node.isRoot())
				{
				   int num_neighbors = node.getNumDirChildren();
				   if (num_neighbors > num_root_neighbors)
				   {
					num_root_neighbors = num_neighbors;
					root = node;
				   }
				}
			}
		}
		
		if (num_nodes == 0)
			return;
		if (root == null && node != null )
			root = node;
	
		Node[] nodes = new Node[num_nodes];	
		int[] parents = new int[num_nodes];
		int[][] children = new int[num_nodes][];
		double[] radiuss = new double[num_nodes];
		int[] depths = new int[num_nodes];
		double[] angle_sizes = new double[num_nodes];

		// Each tree is grow from a single root
		// This will allow for disconnected graphs
		
		int[] trees = new int[num_nodes];
		int current_tree = 0;
		
		// index 0 is saved for the root node
		nodes_unset.reset();
		int index = 1;

		root.setToDefault();

		// Integer.MIN_VALUE is used for undefined
		while (!nodes_unset.end())
		{
			node = (Node) nodes_unset.next();
			if (node == root)
			{
				node.setSize(0);
				nodes[0] = node;
				parents[0] = 0;
				radiuss[0] = 0;
				depths[0] = 0;
				angle_sizes[0] = 2*Math.PI;
				trees[0] = current_tree;
			}
			else
			{
				node.setSize(index);
				nodes[index] = node;
				parents[index] = Integer.MIN_VALUE;
				radiuss[index] = Double.MIN_VALUE;
				depths[index] = Integer.MIN_VALUE;
				angle_sizes[index] = Double.MIN_VALUE;
				trees[index] = Integer.MIN_VALUE;
				index++;
			}
		}
			
		// This creates all of the dependancies
		int root_index = 0;
		do {
			ProcessValueBased(true,root_index, nodes,parents ,
				children, radiuss,
				depths, angle_sizes ,trees, nodes_unset);
			int num_root_children = 0;
			root_index= -1;
			nodes_unset.reset();
			while (!nodes_unset.end())
			{
				index = nodes_unset.next().getSize();
				int num_children= nodes[index].getNumDirChildren();
				if (num_children > num_root_children 
					&& parents[index] != Integer.MIN_VALUE)
				{
					root_index= index;
					num_root_children = num_children;
				}
			} 
			
			if (root_index > 0)
			{
				parents[index] = index;
				radiuss[index] = 0;
				depths[index] = 0;
				angle_sizes[index] = 2*Math.PI;
				trees[index] = ++current_tree;
				root_index = index;
			}
		} while (root_index > 0);
		root_index = 0;


		// This should normallize the radius so they fit on the 
		// screen.
	
		// This gets the max radius
		index = 0;
		root.setToDefault();
		int node_size = root.getSize();
		double max_radius = 0;
		do {
			nodes[index].setSize(node_size);
			double radius = radiuss[index++];
			if (radius > max_radius)
			{
				max_radius = radius;
			}
		} while (index < num_nodes);
		//max_radius *= 2;

		int screen_size= dim.height;
		if (dim.width < screen_size)
			screen_size= dim.width;
		index = 0;
		do {
//System.err.print("  "+nodes[index].getName()+":"+radiuss[index] +"*");
			radiuss[index] = (radiuss[index]*screen_size)
				/max_radius;
//System.err.println(screen_size + "/"+max_radius +"="+radiuss[index]);
			index++;
		} while (index < num_nodes);

		double[] anglesNeeded = new double[num_nodes];
		ShrinkAngle(root_index, size, children, radiuss, angle_sizes, 
			anglesNeeded);
		SetXY(root_index, 0, nodes,children,radiuss,
			angle_sizes);
	}

	protected double[] ProcessValueBased(boolean root,int current, 
		Node[] nodes, int[] parents, int[][] children, double[] radiuss,
		int[] depths, double[] angle_sizes, int[] trees, DList unset)
	{
String test_node= "204.147.130.85";
		Node current_node = nodes[current];
		unset.drop(current_node);
		boolean has_value = current_node.hasValues();
		double value = current_node.getDouble();
		if (value < 0)
			value *= -1;
		//value = Math.log(value);
		Node[] node_children = current_node.getDirChildren();

		int[] mapping = new int[children.length];
		int num_unset_children = 0;

		int depth = depths[current];
		double radius = radiuss[current];
		int tree = trees[current];

		// The amount that this node will have to move to make room
		// for it's children which have small values
		double shift = 0;

		double[] radius_depth = null;

		for (int i=0; i < node_children.length;i++)
		{
			int index = node_children[i].getSize();
			if (parents[index] == Integer.MIN_VALUE)
			{
				mapping[num_unset_children++] = index;
				parents[index] = current;
				depths[index] = depth+1;
				trees[index] = tree;
			}
			else if (radius_depth == null)
			{
				radius_depth = new double[2];
				radius_depth[0] = radiuss[index];
				radius_depth[1] = depths[index];
			}
			else
			{
				double node_radius = radiuss[index];
				if (node_radius < radius_depth[0])
				{
					radius_depth[0] = node_radius;
					radius_depth[1] = depths[index];
				}
			}
		}	
if (current_node.getName().indexOf(test_node) >= 0)
{
	System.err.println(test_node);
	if (radius_depth != null)
		System.err.println("\tcurrent:"+radius_depth[0]+","
			+radius_depth[1]);
	System.err.println("----------");
}
		

		children[current] = new int[num_unset_children];
		double angle_size = angle_sizes[current]/num_unset_children;
//System.err.println("angle_size:"+ angle_sizes[current]+"/"+num_unset_children+"="+angle_size);
		angle_size = angle_sizes[current]/num_unset_children;
		if (has_value || root)
		{
			if (value < Double.MAX_VALUE && value > Double.MIN_VALUE
				&& radius != value && !root)
			{
				radius = value;
				radiuss[current] = value;
				radius_depth = new double[2];
				radius_depth[0] = radius;
				radius_depth[1] = depth;
			}
			if (!root)
				radius += radius/depth;
			else
				radius = 1;
			// Remeber to carry over any changes below
			for (int i=0; i < num_unset_children;i++)
			{
				int index = mapping[i];
				children[current][i] = index;
				node_children[i] = nodes[index];
				angle_sizes[index] = angle_size;
				radiuss[index] = radius;
				ProcessValueBased(false,index, nodes,
					parents , children, radiuss, depths, 
					angle_sizes ,trees, unset);
			}
		}
		else
		{
			radius += radius/depth;
			// Remeber to carry over any changes above
			for (int i=0; i < num_unset_children;i++)
			{
				int index = mapping[i];
				children[current][i] = index;
				node_children[i] = nodes[index];
				angle_sizes[index] = angle_size;
				radiuss[index] = radius;
				double[] node_radius_depth = 
				ProcessValueBased(false,index, nodes,
					parents , children, radiuss, depths, 
					angle_sizes ,trees, unset);
/*
System.out.print(current_node.getName()+":");
if (radius_depth != null)
	System.out.print(radius_depth[0]);
else
	System.out.print("null");
System.out.print("->");
if (node_radius_depth != null)
	System.out.print(node_radius_depth[0]);
else
	System.out.print("null");
System.out.print("\n");
*/
				if (node_radius_depth != null &&
					node_radius_depth[1] > 0 &&
					(radius_depth == null ||
					node_radius_depth[0] < radius_depth[0]))
				{
if (current_node.getName().indexOf(test_node) >= 0)
{
	System.err.println(test_node);
	if (radius_depth != null)
		System.err.println("\tcurrent:"+radius_depth[0]+","
			+radius_depth[1]);
	if (node_radius_depth != null)
		System.err.println("\t  to be:"+node_radius_depth[0]+","
			+node_radius_depth[1]);
}
					radius_depth = node_radius_depth;
				}
			}
			if (radius_depth != null)
			{
                     		radiuss[current] = (radius_depth[0]*depth)
					/radius_depth[1];
				radius_depth[1]--;
			}
		}
if (current_node.getName().indexOf(test_node) >= 0)
	System.err.println("final:"+radiuss[current]);
		current_node.setChildren(node_children);
		return radius_depth;
	}

	protected double ShrinkAngle(int root, int size, int[][] children ,
		double[] radiuss, double[] angle_sizes, double[] anglesNeeded)
	{
		int[] total_children = new int[radiuss.length];
		CountChildren(root,total_children,children);
		ShrinkAngleRecusive(root,total_children,children,angle_sizes);
		return 0;
	}

	protected int CountChildren(int current, int[] total_children,
		int[][] children)
	{
		int total = 1;
		int[] myChildren = children[current];
		for (int i=0; i < myChildren.length;i++)
		{
			int index = myChildren[i];
			total += CountChildren(index, total_children, children);
		}
		total_children[current] = total;	
		return total;
	}

	protected void ShrinkAngleRecusive( int current, int[] total_children,
		int[][] children, double[] angle_sizes)
	{
		int total = total_children[current];
		double angle = angle_sizes[current];
		int[] myChildren = children[current];
		
		for (int i=0; i < myChildren.length;i++)
		{
			int index = myChildren[i];
			angle_sizes[index] =(angle*total_children[index])/total;
			ShrinkAngleRecusive(index, total_children, children,
				angle_sizes);
		}
	}
		
		
/*
	//------------------------- ShrinkAngle  -----------------
	// negative numbers means that this node's angle got smaller
	//	because it didn't need all the space
	// Positive numbers menas that this node's angle needs 
	//	to grow by that amount.
	//--------------------------------------------------------
	protected double ShrinkAngle(int current, int size, int[][] children ,
		double[] radiuss, double[] angle_sizes, double[] anglesNeeded)
	{
System.err.println("ShrinkAngle:" +size);
		int[] my_children = children[current];

		double change = 0;
		if (my_children != null)
		{
		   double shrink = 0;
		   double grow = 0;
		   for (int i=0;i < my_children.length;i++)
		   {
			int index = my_children[i];
			double node_change = ShrinkAngle( size, index, children,
				radiuss, angle_sizes, anglesNeeded);
			if (node_change < 0)
				shrink += node_change;
			else if (node_change < 0)
				grow += change;
		   }
		
		   if (grow > 0)
		   { 
		   	double exchange = 0;
			if (shrink <= -1*grow)
				exchange = grow;
			else
				exchange = -1*shrink;
			for (int i=0; i < my_children.length;i++)
			{
				int index = my_children[i];
				double amount = exchange*anglesNeeded[index]
						/grow;
				if (amount > 0)
					GrowAngle(index, amount, children,
						angle_sizes, anglesNeeded);
			}
		   }
		   change = shrink + grow;
		}

		double wanted_angle = size/radiuss[current];
		double my_change = wanted_angle - angle_sizes[current];
		if (change < my_change)
			change = my_change;
		anglesNeeded[current] = change;
		return change;
	}

	protected void GrowAngle(int current, double size, int[][]children
		, double[] angle_sizes, double[] anglesNeeded)
	{
System.err.println("GrowAngle:" + size);
		if (anglesNeeded[current] < 0)
			return;

		int[] my_children = children[current];
		double needed = anglesNeeded[current];

		for (int i=0; i < my_children.length;i++)
			GrowAngle(my_children[i],size/needed,children,
				angle_sizes, anglesNeeded);

		angle_sizes[current] += size;
		anglesNeeded[current] += size;
	}		
*/
		
	protected void SetXY(int current, double base_angle, 
		Node[] nodes, int[][] children, double[] radiuss,
		double[] angle_sizes)
	{
		double angle = angle_sizes[current]/2;
		double radius = radiuss[current];
		int x = (int)(radius*Math.cos(angle+base_angle));
		int y = (int)(radius*Math.sin(angle+base_angle));
		nodes[current].setXY(x,y);
		
		int[] my_children = children[current];
		for (int i=0;i < my_children.length;i++)
		{
			int index = my_children[i];
			SetXY(index,base_angle,nodes,children,radiuss,
				angle_sizes);
			base_angle += angle_sizes[index];
		}
	}
		
/*********************** Geo Methods ***********************************/
	
	protected void ProcessGeo()
	{
		Dimension dim = display.size();
		DList queue = SetUpGeo();
		DList dlist = display.getList();
		DList ordered = new DList();
		if (queue == null)
			return;

		queue.reset();
		while(!queue.end()) {
			DisplayObject object = queue.next();
			if (((Node)object).getRank()==1)
				ordered.enqueue(object);
		}

		while (!queue.empty())
		{
			Node top = (Node) queue.pop();
			int size = top.getSize();
			int rank = top.getRank();
			Node[] neighbors = top.getNeighbors();
			int num_neighbors = neighbors.length;


			// This removes any at the end 
			while (num_neighbors > 0 
				&& neighbors[num_neighbors-1].isSetXY()) 
				num_neighbors--;
			// This removes the rest
			int where = 0;
			while (where < num_neighbors)
			{
				if (neighbors[where].isSetXY())
				{
				  num_neighbors--;
				  neighbors[where] = neighbors[num_neighbors];
				}
				else
					where++;
			}	
	
			Node[] children = new Node[num_neighbors];
	
			for (int index=0;index<num_neighbors;index++)
			{
			    Node node = neighbors[index];
			    children[index] = node;
			    if (!node.isSetXY())
			    {
				DisplayObject object  = (DisplayObject)node;
			  	double lat = object.getLatitude();
			  	double lon = object.getLongitude(); 
			  	int to_map[] = display.ToMap(lat,lon);
				node.setXY(to_map[0],to_map[1]); 
			        node.setNameLocation(1.5*Math.PI);
				node.setRank(rank+1);

			      	if (size > 3)
					node.setSize(size-1);
			      	else 
			        	node.setSize(1);

			        queue.enqueue(node);

				ordered.enqueue((DisplayObject)node);
			    }
			}
			top.setChildren(children);
		}

		dlist.reset();
		while(!dlist.end())
		{
			DisplayObject object = dlist.next();
			object.setLatLon(true);
		}
		dlist.reset();
		while(!dlist.end())
		{
			DisplayObject object = dlist.next();
			double lat = object.getLatitude();
                        double lon = object.getLongitude();
                        if (object instanceof Node && (lat > 90 || lat < -90 || lon > 180 || lon < -180))
                         {
				object.setLatLon(false);
				DList links = ((Node)object).getLink();
				links.reset();
				while(!links.end())
				{
					object = links.next();
					object.setLatLon(false);
				}
			}
		}
		display.setListNoPaint(dlist);
		display.setOrderedList(ordered);
	}

	protected DList SetUpGeo()
	{
		Dimension dim = display.size();
	
		DList queue = new DList();

		DList dlist = display.getList();
		DList roots = new DList();

		// This is a special case if there are not roots.
		Node most= null;
		int most_size = 0;
		int most_root = 0;
		int most_count = 0;

		dlist.reset();
		while (!dlist.end())
		{
		    DisplayObject object = dlist.next();
		    if (object instanceof Node)
  		    {
			Node node = (Node) object;
			int count = node.getNumNeighbors();
			node.unSetXY();

			if (node.isRoot())
			{
				if (count > most_root)
					most_root = count;
				int size = node.getSize();
				if (size > most_size)
					most_size = size;
				roots.push(node);
			}
			if (count > most_count)
			{
				most = node;
				most_count = count;
			}
		    }
		}

		if (roots.count() <= 0)
			if (most == null)
				return null;
			else
			{
				roots.push(most);	
			}
			
		if (roots.count() == 1)
		{
			Node root = (Node) roots.pop();
			root.setToDefault();
			DisplayObject obj = (DisplayObject)root;
			double lat = obj.getLatitude();
			double lon = obj.getLongitude(); 
			int to_map[] = display.ToMap(lat,lon) ;
			root.setXY(to_map[0],to_map[1]);
			root.setNameLocation(1.5*Math.PI);
			root.setRank(1);

			queue.enqueue(root);
		}
		else
		{
			int count = roots.count();
			for (int index=1; index <= count; index++)
			{
				Node root = (Node) roots.pop();
				root.setToDefault();
				DisplayObject obj = (DisplayObject)root;
				double lat = obj.getLatitude();
				double lon = obj.getLongitude(); 
				int to_map[] = display.ToMap(lat,lon);
				root.setXY(to_map[0],to_map[1]);
				root.setNameLocation(1.5*Math.PI);
				root.setRank(1);
				queue.enqueue(root);
			}
		}
		return queue;
	}

/*************************************************************************/

/*********************** Pull and Push ***********************************/

 	protected void ProcessPullPush()
	{
		DList dlist = display.getList().DList();
		DList node_list = new DList();
		DList link_list = new DList();
		dlist.reset();
		DisplayObject object = null;
		int num_links =0;
		int num_nodes =0;
		int size = 0;

		// I use the size of the node to show it's location 
		// in the array.
		while (!dlist.end())
		{
			object = dlist.next();
			if (object instanceof Node)
			{
				node_list.enqueue(object);
				num_nodes++;
			}
			else if (object instanceof Link)
			{
				num_links++;
				link_list.enqueue(object);
			}
		}
		
		//node_list = MoveLeafsToEnd(node_list);

		// Set up all the list 
		Node[] nodes = new Node[num_nodes];
		//double[] x = new double[num_nodes];
		//double[] y = new double[num_nodes];

		double[] x_v = new double[num_nodes];
		double[] y_v = new double[num_nodes];
		double[] v = new double[num_nodes];

		int[][] links = new int[num_links][2];

		link_list.reset();
		int link_index = 0;
		while (!link_list.end())
		{
			Link link = (Link) link_list.next();
			links[link_index][0] = link.getChild().getSize();
			links[link_index][1] = link.getParent().getSize();
			link_index++;
		}
		
		node_list.reset();
		int node_index =0;
		while (!node_list.end())
		{
			Node node = (Node) node_list.next();
			node.setToDefault();
			nodes[node_index++] = node;
		}
		double max_change = 0;
		int on_top = 0;
		int times = 20;
		size = SIZE;
		do {
			//  Setting every to 0
			for (int i=0;i < num_nodes; i++)
			{
				x_v[i] = 0;
				y_v[i] = 0;
				v[i] = 0;
			}

			//   Push
			for (int i=0;i < num_nodes; i++)
			{
			 for (int j=i+1;j < num_nodes; j++)
			 {
				double delta_x = nodes[i].getX() - nodes[j].getX();
				double delta_y = nodes[i].getY() - nodes[j].getY();
				if (delta_x == 0 && delta_y == 0)
				{
					switch(on_top)
					{
					    case 0:
						delta_x = 0;
						delta_y = 1;
						break;
					    case 1:
						delta_x = 0;
						delta_y = -1;
						break;
					    case 2:
						delta_x = 1;
						delta_y = 1;
						break;
					    case 3:
						delta_x = -1;
						delta_y = -1;
						break;
					    case 4:
						delta_x = 1;
						delta_y = 0;
						break;
					    case 5:
						delta_x = -1;
						delta_y = 0;
						break;
					    case 6:
						delta_x = 1;
						delta_y = -1;
						break;
					    case 7:
						delta_x = -1;
						delta_y = -1;
						break;
					}
					nodes[j].changeXY((int)delta_x,(int)delta_y);
				 	on_top = (on_top+1)%8;
				}
				double delta_v = delta_x*delta_x +delta_y*delta_y;
				double change_v = PUSH/delta_v; 
				delta_v = Math.sqrt(delta_v);
				double change_x = (change_v*delta_x)/delta_v;
				double change_y = (change_v*delta_y)/delta_v;
				
				x_v[i] += change_x; 
				y_v[i] += change_y; 
				x_v[j] -= change_x; 
				y_v[j] -= change_y; 
			 }
			}
			
			for (int link =0; link < num_links; link++)
			{
				int i= links[link][0];
				int j= links[link][1];
				double delta_x =nodes[i].getX()-nodes[j].getX();
				double delta_y =nodes[i].getY()-nodes[j].getY();
				double delta_v =delta_x*delta_x+delta_y*delta_y;
				double change_v = PULL*delta_v; 
//System.err.println("change_v:"+delta_x+"*"+nodes[i].getY()+","+nodes[j].getY()+","+delta_y+"="+change_v);
				delta_v = Math.sqrt(delta_v);
				double change_x = (change_v*delta_x)/delta_v;
				double change_y = (change_v*delta_y)/delta_v;
				
				x_v[i] -= change_x; 
				y_v[i] -= change_y; 
				x_v[j] += change_x; 
				y_v[j] += change_y; 
			}

			max_change = 0;
			for (int i=0; i < num_nodes; i++)
			{
				v[i] = x_v[i]*x_v[i] + y_v[i]*y_v[i];
				if (v[i] > max_change)
					max_change = v[i];
			}
		
			if (max_change > MAX_GROWTH*size*size)
				max_change /= MAX_GROWTH*size*size;
			else
				max_change = 1;

			double scaler = 1/Math.sqrt(max_change);
			for (int i=0; i < num_nodes; i++)
			{
				if (scaler > 0)
				{
					int x_scaler = (int)(x_v[i]*scaler);
					if (x_scaler > MAX_GROWTH)
						x_scaler = (int)MAX_GROWTH;
					int y_scaler = (int)(y_v[i]*scaler);
					if (y_scaler > MAX_GROWTH)
						y_scaler = (int)MAX_GROWTH;
					nodes[i].changeXY(x_scaler,y_scaler);
				}
			}
			
		times--;
		} while (times > 0);
		for (int i=0; i < num_nodes; i++)
			nodes[i].setToDefault();
	}
	
/*************************************************************************/

/*********************** Force Methods **********************************/
	
/*
	// The number of cycles to go throught
	int num_cycles = 1;
	int num_steps  = 1;

	// Node Wight
	double nodeWieght = 10;
	double neighborWieght = 1;

	// An array of all the nodes
	Node[] nodes;
	// An array of all the link pairs
	Node[][] links;

	protected void ProcessForce()
	{
		SetUpForce();
		if (nodes.length > 0)
			size = 4*nodes[0].getSize();
		int cycles = num_cycles * num_steps; 
		while (cycles > 0)
		{
			if (cycles %num_steps == 0)
				printMessage("Number of steps left " 
					+ cycles);
			for (int index=nodes.length-1;index >= 0; index--)
			{	
				Node node = nodes[index];
				// push works on both nodes so it only needs to						// be called once for each par
				for (int i=index-1; i >=0;i--)
					node.push(nodes[i]);
			}
			for (int index=links.length-1;index >=0;index--)
				links[index][0].pull(links[index][1]);
			for (int index=nodes.length-1;index >= 0; index--)
				nodes[index].change();
			cycles--;
		}
	}

	protected void SetUpForce()
	{
		DList dlist = display.getList();
		dlist.reset();
		int num_nodes = 0;
		int num_links = 0;
		while(!dlist.end())
			if (dlist.next() instanceof Node)
				num_nodes++;
			else 
				num_links++;
		nodes = new Node[num_nodes];
		links = new Node[num_links][2];
		dlist.reset();
		num_nodes = 0;
		num_links = 0;
		while(!dlist.end())
		{
			DisplayObject object = dlist.next();
			if (object instanceof Node)
			{
				nodes[num_nodes] = (Node)object;
				num_nodes++;
			}
			else if (object instanceof Link)
			{
				links[num_links][0] = (Node) ((Link)object).getParent(); 
				links[num_links][1] = (Node) ((Link)object).getChild();
				num_links++;
			}
		}
	}
*/
}
