package caida.otter;

import java.io.*;
import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.*;
import java.net.*;
import java.net.URL;
import java.util.Date;

import java.util.StringTokenizer;

import caida.tools.PsGenerator;

/**************************************************************************
* File: Otter.java		Name: Otter
* Goal: This applet/application provides a interactive graphically represent 
*	of a network.  Where that network can have a file specified type
*	and number of attributes for each data object.  It will provide
*	the a GUI for the user.
*
*	Allow user to load a network
*	Allow user to save a network if in stand alone
*	Allow user to create, remove, and  change the location of nodes and 
*             links that have already be created
*
*	As a applet it will need a data_address parameter.
*
* Written: Bradley Huffaker
* For: Cooperative Association for Internet Data Analysis
***************************************************************************
***************************************************************************
By accessing this software, OTTER, you are duly informed of and 
agree to be bound by the conditions described below in this notice:
 
This software product, OTTER, is developed by Bradley Huffaker, and
copyrighted(C) 1998 by the University of California, San Diego (UCSD),
with all rights reserved.  UCSD administers the NSF grant to CAIDA,
number NCR-9711092, under which this code was developed. 
 
There is no charge for OTTER software. You can redistribute it and/or
modify it under the terms of the GNU General Public License, v. 2 dated
June 1991 which is incorporated by reference herein.  OTTER is
distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that the use of
it will not infringe on any third party's intellectual property
rights.
 
You should have received a copy of the GNU GPL along with the OTTER
program.  Copies can also be obtained from
http://www.gnu.org/copyleft/gpl.html or by writing to
 
                        University of California, San Diego
                        SDSC/CAIDA
                        9500 Gilman Dr., MS-0505
                        La Jolla, CA 92093 - 0505  USA
 
  Or contact INFO@CAIDA.ORG
**************************************************************************/    
public class Otter extends Applet
{

    String address = null;

    protected int width = 700;
    protected int height = 500; 

    //protected int width = 600;
    //protected int height = 450; 
    
    public static boolean isApplet = true;
    public static boolean hasWindow = true;
    public static boolean showAllNames = false;
    
    OtterFrame frame;
    /**Command line params...*/
    // The file that is autoloaded if set
    static String auto_load_file = null;

    int colorByGroup=-1; //for the -g param
    int colorByField=-1; 

    int nameForGroup; //for the -n param
    int nameForField; 
    String nameForConditional; 

    String outputFile;
    int outputWidth;
    int outputHeight;
   
    Button control;
    
    //jjung Map stuff;
    Image image;

 /*********************************************
  * Gives some usefule info about this applet *
  *********************************************/
    
 public String getAppletInfo()
 {
	return "Otter - Developed by: Bradley Huffaker";
 }

 /***************************************************
  * This will set up the system if run as a applet. *
  ***************************************************/

 public void init()
 {

	//jjung

        Arrows.init(16, 0.5);
        setFont(new Font("Serif", Font.PLAIN, 10));

	String image_path = "../Images/world1.gif";
	if (isApplet)
	{
	    setBackground(Color.white);
	    address = getParameter("data_address");
	    if (address != null) 
		address = this.getCodeBase().toString() + address;
	    //width  = Integer.parseInt(getParameter("frame_width"));
	    //height = Integer.parseInt(getParameter("frame_height"));
	    String image_address = getParameter("image_address");;
	    
	    if (image_address != null)
		{
		    int http_found = image_address.indexOf("http://");
		    if (http_found < 0)
			image_address = getCodeBase().toString()
			    +"/" + image_address;
		    try {
			image = this.getImage(new URL(image_address));
		    } catch (MalformedURLException e0) {
			System.err.println(e0.toString()); }
		}
	    
	    // This is a auto load file
	    auto_load_file = getParameter("auto_load_file");
	    
	    control = new Button("Load and View");
	    add(control);
	}
	//else if (outputFile!=null){ //we don't wanna put anything on the screen
	//}
	else
	{
	    image = getToolkit().getImage(image_path);
	    //System.out.println("image"+ image_path + "is fetched\n");
	    frame = new OtterFrame(isApplet,address,auto_load_file,
				   image,width,height);
	    //stop();
	}
 }

 public boolean action (Event event, Object arg)
 {
     if (event.id == Event.ACTION_EVENT && event.target == control)
	 {
	     frame = new OtterFrame(isApplet,address,auto_load_file
				    ,image,width,height);
	 }
     return super.action(event,arg);
 }

    /**Extracts command line arguments of form -Xopt1,opt2,...,optN.
       @return array of form [X,opt1,opt2,...,optN], null on error
    */
    private Object[] extractCommaDelimetedArguments(String args)
    {
	if (args.charAt(0)!='-')
	    return null;
	StringTokenizer st=new StringTokenizer(args.substring(2),",");
	Object[] ret=new Object[st.countTokens()+1];
	ret[0]=new Character(args.charAt(1));
	for (int i=1;i<ret.length;i++){
	    ret[i]=st.nextToken();
	}

	for (int i=1;i<ret.length;i++){
	    System.out.println("\t["+i+"]="+ret[i]);
	}
	return ret;
    }

	
    /**
       Handles the command line. Outputs a description of usage on error.
       @return true if processing successful. false otherwise.
    */
    private boolean processCommandLineArguments(String[] args)
    {
	boolean error=false;
	if (args.length==0) return true ;

	//First make sure the file specified as the last argument acutally exists and is readable. 
	//Otherwise, there isn't a point to parsing the rest of the args

	File inputFile=new File(args[args.length-1]);
	try{
	    if (!inputFile.canRead()){
		System.err.println("Could not read input file specified: "+args[args.length-1]);
		error=true;
	    }
	}catch (SecurityException e){
	    System.err.println("Could not read input file specified. It seems that a SecurityManager is stopping me."+e);
	    error=true;
	}
	auto_load_file=args[args.length-1];


	int index = 0;
	while (index < (args.length -1) && !error) // the -1 is because we don't process the last arg
	{
		if (args[index].charAt(0) == '-')
		{
		    Object[] command=extractCommaDelimetedArguments(args[index]);
		    if (command==null){
			error=true;
			break;
		    }

		    switch (((Character)command[0]).charValue()){
		    case 'g':
			if (command.length!=3){
			    error=true;
			    break;
			}
			    
			try{
			    colorByGroup=(new Integer((String)command[1])).intValue(); //todo: check these things
			    colorByField=(new Integer((String)command[2])).intValue();
			}catch(Exception e){
			    error=true;
			    System.err.println("ERROR: -g expects integral options. Maybe you mistyped???");
			}
			
			break;
		    case 'n':
			if (command.length!=4){
			    error=true;
			    System.err.println("ERROR: the -n parameter expects three options");
			    break;
			}
			try{
			    nameForGroup=(new Integer((String)command[1])).intValue();
			    nameForField=(new Integer((String)command[2])).intValue();
			    nameForConditional=(String)command[3]; //maybe the parsing code should happen elsewhere
			}
			catch(Exception e){
			    error=true;
			    System.err.println("ERROR: options to the -n parameter were invalid. It expected ints and got something else.");
			}
			break;
		    case 'd':
			if (command.length!=3){
			    error=true;
			    System.err.println("ERROR: the -d parameter expects two options");
			    break;
			}
			try{
			    outputWidth=(new Integer((String)command[1])).intValue();
			    outputHeight=(new Integer((String)command[2])).intValue();
			}catch(Exception e){
			    error=true;
			    System.err.println("Sorry, but the dimensions you specified must be integral.");
			}
			break;
                    case 's': 
                        if (command.length!=2) {
                            error=true;
                            System.err.println("ERROR: the -s parameter expects true or false");
                            break;
                        }
                        try {
                            showAllNames = new Boolean(((String)command[1])).booleanValue();
                        } catch (Exception e) {
                            error=true;
                            System.err.println("ERROR: the -s parameter expects a true or false");
                        }
                        break;

		    case 'o':
		        if (command.length!=2){
			     error=true;
			     System.err.println("ERROR: no save file specified");
			     
			     break;
			}
			//make sure we can actually use the file given
			File outFile=new File((String)command[1]);
			try{
			    if (outFile.exists()&&!outFile.canWrite()){
				System.err.println("Could not write to output file specified: "+command[1]);
				error=true;
			    }
			    //uncomment for jdk>1.2
			    /*else if (!outFile.exists() && !outFile.createNewFile()){
				System.err.println("Could not create output file specified: "+command[1]);
				error=true;
			    }*/
				    
			}catch (SecurityException e){
			    System.err.println("Could not write to output file specified. It seems that a SecurityManager is stopping me."+e);
			    error=true;
			}
			outputFile=(String)command[1];
                        hasWindow = false;
			break;
		    
		    default:
			error=true;
			break;
		    }
		}
		else
		    error = true;
		index++;
	}
	
	
	if (error) printCommandLineArgumentsHelp();
	return !error;
    }

    private void printCommandLineArgumentsHelp()
    {
	System.err.println("Otter - provides an interactive graphical representation of a network\n");
	System.err.println("otter [-gG,A] [-nG,A,conditional] [-dX,Y] [-o outputfile] [inputfile]");
	System.err.println("\t-g\tG=color by Group (0/1), A=color by field (0/1)");
//	System.err.println("\t-n\tG=name for Group, A=name for field, threshold");
	System.err.println("\t-d\tWidth and Height to render");
	System.err.println("\t-o\tFile to output to");
        System.err.println("\t-s\tShow all names");
	System.err.println("\nSee http://www.caida.org for a more detailed description of this tool.");
	//System.err.println("\tinputfile\tData description file");
    }


    public static void main(String arg[])
    {
	isApplet=false;
	Otter otter = new Otter();
	if (!otter.processCommandLineArguments(arg))
	    System.exit(1);
	otter.init();
    }


// This is the frame itself
class OtterFrame extends Frame implements IOInterface, FileInterface
{

 	// Handle most of the mode changeing events
 	// Extends Menu Bar
  protected PullDownMenu menu;
	// Handle the drawing and node/link events
	// Extends Canvus
  protected Display display;
	// Handle the auto formatting of the nodes
  protected Format format;
	// Handle the File IO
  protected IOHandler ioHandler;
	// Remeber if it is a applet or not
  protected boolean isApplet;
	// The address of the data file if this is an applet.
  protected String address;
	// This is the path to the directory where the data file is.
  protected String dir;

	// The scroll bars varaibles
  protected Scrollbar hbar;
  protected Scrollbar vbar;
  final protected int SCROLLBAR_SIZE = 100;
	// The pop up frame that shows help info
  protected HelpFrame help;
	// The pop up frame that does the name search
  protected NameSearch nameSearch;
	// The pop up frame that shows step by step
  protected StepByStep stepBystep;
	// The pop up frame that does animation between files
  protected Animator animator;
	// The pop up frame that does a URL query for fvl file names
  protected URLDialog urlDialog;
  protected FileDialog fileDialog;
  protected FileDialog imageDialog;
  protected DimensionDialog dimensionDialog;
  //jjung
  Image image;

  protected ValuesGroup[] groups = new ValuesGroup[0];

  public OtterFrame(boolean isApplet_input, String address_input, 
	String auto_load_file, Image image_input, int width, int height)
  {
   	super("Otter: a network mapping tool");
	isApplet = isApplet_input;
	address = address_input;
	image = image_input;
	dir = "Data";
   	resize(width,height);

	menu = new PullDownMenu(isApplet);

	display = new Display(image, 90,-180,-90,180);
	display.setMode(Display.SELECT);
	if (outputFile!=null) display.setOffScreenMode(true);
	format = new Format(display);

	help = new HelpFrame(HelpFrame.ABOUT);
	nameSearch = new NameSearch(display);
	stepBystep = new StepByStep(display);

/*
	hbar = new Scrollbar(Scrollbar.HORIZONTAL, SCROLLBAR_SIZE/2, 0, 0, 
			SCROLLBAR_SIZE);
	vbar = new Scrollbar(Scrollbar.VERTICAL, SCROLLBAR_SIZE/2, 0, 0, 
			SCROLLBAR_SIZE);
	vbar = new java.awt.Scrollbar(Scrollbar.VERTICAL);
+*/

	//------ Object placement--------------------------
	setMenuBar(menu);	
	
	setLayout(new BorderLayout());
	add("Center",display);

	addWindowListener(new WindowAdapter(){
	    public void windowClosing(WindowEvent e){
		System.exit(0);
	    }
	});


	if (outputFile==null) //we don't wanna be visible if we are going straight to disk
	    show();
	else 
	    display.setSize(698,445);

	if (auto_load_file != null)
		StartLoading(auto_load_file);
	else if (isApplet)
		StartLoading();
	//if (outputFile==null)
	display.setRes(Display.LOW);

 }

 synchronized public void printMessage(String msg)
 {
	display.printMessage(msg);
 }
 
	/********************************************************
	* FinishedLoading & FinishedSaving
	* Purpose to provide a common interface for reading and 
	* saving files.
	*********************************************************/
 boolean ioHandled = false;
 public void StartLoading()
 {
	StartLoading(null);
 }
 public void StartLoading(String auto_load_file)
 {
        // Don't paint untill we have all the links and nodes
        display.ready = false;
 	ioHandled = false;

        printMessage("Loading ... Please wait.");
	
	if (isApplet)
	{
		if (ioHandler != null)
			ioHandler.stop();
		ioHandler = new IOHandler(this);
		Dimension dim = size();
		ioHandler.setStart(dim.width/2,dim.height/2);

		String file_address = auto_load_file;
		if (file_address == null)
		{
			if (urlDialog == null)
				urlDialog = new URLDialog(this,address);
			urlDialog.Show();
			file_address = urlDialog.getAddress();
		}
		//else
		    //			file_address = file_address;
		if (!ioHandled)
		{
			if (file_address != null)
			{
		   		ioHandled = true;
                                printMessage("Loading ... Please wait.");
		   		ioHandler.Load(file_address);
			}
			else
				printMessage("aborted");
		}
	}
	else if (dir != null || auto_load_file != null)
	{
	    if (ioHandler != null)
		ioHandler.stop();
	    ioHandler = new IOHandler(this);
	    Dimension dim = size();
	    ioHandler.setStart(dim.width/2,dim.height/2);
	    
	    if (auto_load_file == null)
		{
		    if (fileDialog == null)
			fileDialog = new FileDialog(this
						    ,"Load a Data File", FileDialog.LOAD);
		    fileDialog.setDirectory(dir);
		    fileDialog.setFilenameFilter(new NameFilter());
		    fileDialog.show();
		    String file = fileDialog.getFile();
		    dir = fileDialog.getDirectory();
		    if (file != null) {
                        printMessage("Loading ... Please wait.");
			ioHandler.Load(file,dir);
		    } else {
			printMessage("");
                    }
		}
	    else {
                printMessage("Loading ... Please wait.");
		ioHandler.Load(auto_load_file,"");
            }
	}
	else 
	    printMessage("The directory you gave was null");

        display.ready = true;
 }
 // Used by URLDialog
 public void setFile(String file)
 {
	printMessage("Loading ... Please wait.");
        display.ready = false;
	if (file != null && !ioHandled)
	{
		if (isApplet)
			ioHandler.Load(file);
		else
			ioHandler.Load(file,dir);
	}
	ioHandled = true;
        display.ready = true;
 }


 // Used by both FilesSelect
 public boolean isApplet() { return isApplet; }
 public void setFiles(String[] strings)
 {
 }
 synchronized public void FinishedLoading(DList list
	, ValuesGroup[] groups_input, int file_format, Date[] dates)
 {
	printMessage("Loading ... Complete.");
	display.setMode(Display.SELECT);
	display.setListNoPaint(list);
	display.setNumFiles(1);
	display.setFileIndexNoPaint(0);
	format = new Format(display);


	if (file_format == IOHandler.FIXED)
	{
		/*
		DList mylist = list.DList();
		mylist.reset();
		while (!mylist.end())
		{
			DisplayObject object = mylist.next();
			if (object instanceof Node)
				((Node)object).setXY();
		}		
		*/
		format.setMode(Format.FIXED);
	}
	else
	{
		if (file_format == IOHandler.GEO)
			format.setMode(Format.SITE1);
		else
			format.setMode(Format.ROOT);
	}

        if (!isApplet && outputFile != null) {
            format.run();
        } else {
// This screws up the screen because while we're still reading input
// file, it start drawing and thus assigning numbers to links while
// we don't know yet how many and where all the links are going to be
//            format.start();
            format.run();
        }

	display.setMap(false);
	display.setResNoPaint(Display.LOW);
	display.setShowMap(false);
	display.setViewGeo(false);
	
	if (groups != null)
	{
		int number = 0;
		for (int index=0 ; index< groups_input.length;index++)
			if (groups_input[index] != null)
				number++;
		groups = new ValuesGroup[number];
		for (int index=0 ; index< groups_input.length;index++)
			if (groups_input[index] != null)
				groups[index] = groups_input[index];
		menu.setGroups(groups);
	}

	if (dates != null)
	{
		display.setDate(dates[0]);
		display.printDate();
	}

	if ((colorByGroup>=0)&&(colorByField>=0)){ //color by grouping specified
	    ValuesGroup group=menu.groups[colorByGroup];
	    group.setSelected(colorByField);
	    display.setGroup(group);
	}

        if (showAllNames) {
            display.showAllNames();
        }

	if (outputFile!=null){ //we have gotten the command line param to save to a file
	    SaveImage(false,outputFile);
	    if (!isApplet)
		System.exit(0);
	}


 }

 synchronized public void FinishedSaving()
 {
 }

 public boolean handleEvent(Event event)
 {
	switch (event.id)
	{
		case Event.ACTION_EVENT:
			if (event.arg instanceof String)
			{
				format.stop();
				String label = (String)event.arg;
			// File Menu
				if (label.equals(PullDownMenu.LOAD))
					StartLoading();
				else if (label.equals(PullDownMenu.LOAD_MAP)){
					if(!display.isSetMap())
						printMessage("Sorry, data contain no geographical information"); 
					else if(!display.isViewGeo())
						printMessage("It doesn't need Map"); 
					else {
						display.setShowMap(true);
						double[] latlon = 
							display.getLatLon();
						display.zoom(latlon[1],latlon[0]
							,latlon[3],latlon[2]);
						display.repaintNodes();
						printMessage("Finished"); 
					}
				}
				else if (label.equals(PullDownMenu 
						.IMAGE_WITH_KEY))
					SaveImage(true);
				else if (label.equals(PullDownMenu
						.IMAGE_WITHOUT_KEY))
					SaveImage(false);
				else if (label.equals(PullDownMenu.CLEAR_MAP))
				{
					display.setShowMap(false);
					display.repaintNodes();
				}
				else if (label.equals(PullDownMenu.CLEAR))
					display.Clear();
				else if (label.equals(PullDownMenu.CENTER))
				{
					display.recenter();
					display.repaintNodes();
				}
				else if (label.equals(PullDownMenu.ANIMATE))
				{
				  String where = "";
				  if (isApplet)
				  	where = address;
				  else
				  	where = dir;
				  if (animator == null)
				  	animator = new Animator(this
					,display,isApplet,where);
				  animator.show();
				}
				else if (label.equals(PullDownMenu.EXIT))
				{
					hide();
					dispose();
					if (!isApplet)
						System.exit(0);
				}

			// Build Menu
				else if (label.equals(PullDownMenu.ADD_NODE))
					display.setMode(Display.ADD_NODE);
				else if (label.equals(PullDownMenu.ADD_LINK))
					display.setMode(Display.ADD_LINK);
				else if (label.equals(PullDownMenu.REMOVE))
					display.setMode(Display.REMOVE);

			// Interaction Menu
				else if (label.equals(PullDownMenu.SELECT))
					display.setMode(Display.SELECT);
			       else if (label.equals(PullDownMenu.SELECT_TREE))
					display.setMode(Display.SELECT_TREE);
				else if (label.equals(PullDownMenu.DESELECT))
					display.deselectAll();
			       else if (label.equals(PullDownMenu.SELECT_NAME))
					nameSearch.show();
			       else if (label.equals(PullDownMenu.ROOT_SELECT))
					display.setMode(Display.ROOT_SELECT);
				else if (label.equals(PullDownMenu.MOVE_NAMES))
					display.setMode(Display.MOVE_NAMES);
				else if (label.equals(PullDownMenu.MOVE))
					display.setMode(Display.MOVE);
				else if (label.equals(PullDownMenu.STEP_BY_STEP)) {
					display.setMode(display.STEP_BY_STEP);
					//display.stepBystep(0);
					stepBystep.show();
				}

			// View Format
				else if (label.equals(PullDownMenu.FORMAT_IP))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.IP);
					format.start();
					//display.repaintNodes();
				}
				else if (label.equals(PullDownMenu.FORMAT_ROOT))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.ROOT);
					format.start();
					//display.repaintNodes();
				}
				else if (label.equals(PullDownMenu.FORMAT_GEO))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.SITE1);
					format.start();
				}
				else if (label.equals(PullDownMenu.FORMAT_FORCE))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.FORCE);
					format.start();
				}
				else if (label.equals(PullDownMenu.FORMAT_VALUE))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.VALUE);
					format.start();
				}
				else if (label.equals(PullDownMenu.FORMAT_FIND))
				{
					display.setShowMap(false);
					display.setViewGeo(false);
					display.setResNoPaint(Display.LOW);
					format = new Format(display);
					format.setMode(Format.FIND_ROOTS);
					format.start();
				}
				else if (label.equals(PullDownMenu.FORMAT_IP_PULL))
				{
					display.ipPull();
					display.repaintNodes();
				}
				else if (label.equals(PullDownMenu.FORMAT_AGIEN))
				{
					if(!display.isViewGeo()) {
						int mode = Format.ROOT;
						if (format != null)
						     mode = format.getMode();
						format = new Format(display);
						format.setMode(mode);
						format.start();
					}
				}

			// View Menu
				else if (label.equals(PullDownMenu.ALL_NAMES))
					display.showAllNames();
				else if (label.equals(PullDownMenu.ROOT_NAMES))
					display.showRootNames();
				else if (label.equals(PullDownMenu.NO_NAMES))
					display.hideNames();
				else if (label.equals(PullDownMenu.LOW))
					display.setRes(Display.LOW);
				else if (label.equals(PullDownMenu.HIGH))
					display.setRes(Display.HIGH);
				else if (label.equals(PullDownMenu.SHOW_DATE))
					display.swapDateShow();
                                else if (label.equals(PullDownMenu.SHOW_ARROWS))
                                        display.swapDrawArrows();

			// Tool Menu
				else if (label.equals(PullDownMenu.DIMENSION))
 					setDimenions();
			// Zoom Menu
				else if (label.indexOf(PullDownMenu.ZOOM) >= 0)
				{
				  // Prase out the part that is a number
				  char[] array = label.toCharArray();
				  int start = 0;
				  while (start < array.length
  				   && (array[start] < '0'
				   || array[start] > '9'))
				    start++;
			           int end = start;
				  while (end < array.length
				   && ((array[end] >= '0'
				   && array[end] <= '9')
				   || array[end] == '.'))
				    end++;
				  if (start < end)
				  {
					
				    double zoom= (new Double(
					new String(array,start,end-start))
					).doubleValue();
				    if(display.isViewGeo())
					display.setMode(display.ZOOM);
				    display.scale(zoom);
				  }
				}
				// Node size Menu
				else if (label.indexOf(PullDownMenu.NODE) >= 0)
				{
				  // Prase out the part that is a number
				  char[] array = label.toCharArray();
				  int start = 0;
				  int minus = 1;
				  while (start < array.length
  				   && (array[start] < '0'
				   || array[start] > '9' ))
				    start++;
				  if(start >0)
					if(array[start-1] =='-')
						minus = -1;
			           int end = start;
				  while (end < array.length
				   && ((array[end] >= '0'
				   && array[end] <= '9')
				   || array[end] == '.'))
				    end++;
				  if (start < end)
				  {
					
				    double node_size= (new Double(
					new String(array,start,end-start))
					).doubleValue();
				    node_size = minus * node_size; 
				    display.nodeSize((int)node_size);
				  }
				}
				// Line size Menu
				else if (label.indexOf(PullDownMenu.LINE) >= 0)
				{
				  // Prase out the part that is a number
				  char[] array = label.toCharArray();
				  int start = 0;
				  int minus = 1;
				  while (start < array.length
  				   && (array[start] < '0'
				   || array[start] > '9' ))
				    start++;
				  if(start >0)
					if(array[start-1] =='-')
						minus = -1;
			           int end = start;
				  while (end < array.length
				   && ((array[end] >= '0'
				   && array[end] <= '9')
				   || array[end] == '.'))
				    end++;
				  if (start < end)
				  {
					
				    double line_size= (new Double(
					new String(array,start,end-start))
					).doubleValue();
				    line_size = minus * line_size; 
				    display.lineSize((int)line_size);
				  }  
				}
                                // Arrow Size Menu
                                else if (label.indexOf(PullDownMenu.ARROW) >=0)
                                {
                                    char[] array = label.toCharArray();
                                    if (array[6] == '-') {
                                        display.decreaseArrowSize();
                                    } else {
                                        display.increaseArrowSize();
                                    }
                                }
                                // Arrow Angle menu
                                else if (label.indexOf(PullDownMenu.ARROW_S)>=0)
                                {
                                    char[] array = label.toCharArray();
                                    if (array[6] == '-') {
                                        display.decreaseArrowAngle();
                                    } else {
                                        display.increaseArrowAngle();
                                    }
                                }
			// Color Method
				else if (label.equals(
						PullDownMenu.METHOD_DEFAULT))
				   display.setMethod(Display.METHOD_DEFAULT);
				else if (label.equals(
						PullDownMenu.METHOD_SPECTRUM))
				   display.setMethod(Display.METHOD_SPECTRUM);
				else if (label.equals(
						PullDownMenu.METHOD_MASK))
				   display.setMethod(Display.METHOD_MASK);
					
			// Help Menu
				else if (label.equals(PullDownMenu.COMMANDS))
				{
					help.setType(HelpFrame.COMMANDS);
					help.toFront();
					help.show();
				}
				else if (label.equals(PullDownMenu.ABOUT))
				{
					help.setType(HelpFrame.ABOUT);
					help.toFront();
					help.show();
				}
				else if (label.equals(PullDownMenu.OVERVIEW))
				{
					help.setType(HelpFrame.OVERVIEW);
					help.toFront();
					help.show();
				}
				else if (label.equals(
					PullDownMenu.METHOD_SPECTRUM))
					display.colorBySpectrum();
				else
				{
 				   if (event.target instanceof MenuItem)
				   {
					ValuesGroup group = 
					  menu.getGroup((MenuItem)event.target);
					if (group != null)
						display.setGroup(group);
				   }
				}
			}
			break;
/*
		default:
		if (event.target == hbar)
			display.Shift_X( ((Integer)event.arg).intValue()/SCROLLBAR_SIZE);
		else if (event.target == vbar)
			display.Shift_Y( ((Integer)event.arg).intValue()/SCROLLBAR_SIZE);
*/
	}
	return super.handleEvent(event);
 }
 
 //**********************************************************************
 // Used to save the backgound image to a file
 //**********************************************************************
 void SaveImage(boolean with_key)
 {
	if (imageDialog == null)
		imageDialog = new FileDialog(this,"Save Network Image"
			,FileDialog.SAVE);
	imageDialog.show();
	String dir = imageDialog.getDirectory();
	String filename = imageDialog.getFile();
	if (dir != null && filename != null)
	{
		filename = dir + filename;
		SaveImage(with_key,filename);
	}
 }

 void SaveImage(boolean with_key, String filename)
 {
     display.printMessage("Saving image...");
     
     PsGenerator postscript = new PsGenerator(filename);
     postscript.setFont(new Font("Serif", Font.BOLD, 10));
     
     Rectangle coords = display.getImageExtremeCoords();
     display.drawPostscript(postscript,with_key,hasWindow);
     //postscript.setColor(Color.red);
     //postscript.fillRect(10,10,40,40);
     
     postscript.pack();
     display.printMessage("Finished Image");
     display.printMessage("");
 }

 //********************************************************************
 // CReates a popup window which will be used to collec the dimensions
 // The users wishs to uses on the image
 //********************************************************************
 void setDimenions()
 {
	if (dimensionDialog == null)
		dimensionDialog = new DimensionDialog(display);
	dimensionDialog.show();
 }

    
} //inner class: OtterFrame
} //class: Otter
    
