package caida.otter;
import java.awt.*;
import java.util.Random;
import java.lang.Math;

/**************************************************************************
* File: Path.java		Class: Path
* Goal: A Display Object that is used to display paths as a collection
*	of lines between nodes.  
*
* Written: Bradley Huffaker             (12/17/97)
*
* Modified: Jaeyeon Jung		(2/13/93)
* getDomain: IP address which doesn't have domain name will
*	     be returned as unknown 
*
* For:Cooperative Association for Internet Data Analysis
***************************************************************************
***************************************************************************
By accessing this software, PATH, you are duly informed of and 
agree to be bound by the conditions described below in this notice:
 
This software product, PATH, is developed by Bradley Huffaker and 
Jaeyeon Jung, and copyrighted(C) 1998 by the University of California, 
San Diego (UCSD), with all rights reserved.  UCSD administers the NSF 
grant to CAIDA, number NCR-9711092, under which this code was developed. 
 
There is no charge for PATH software. You can redistribute it and/or
modify it under the terms of the GNU General Public License, v. 2 dated
June 1991 which is incorporated by reference herein.  PATH is
distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that the use of
it will not infringe on any third party's intellectual property
rights.
 
You should have received a copy of the GNU GPL along with the PATH
program.  Copies can also be obtained from
http://www.gnu.org/copyleft/gpl.html or by writing to
 
                        University of California, San Diego
                        SDSC/CAIDA
                        9500 Gilman Dr., MS-0505
                        La Jolla, CA 92093 - 0505  USA
 
  Or contact INFO@CAIDA.ORG
**************************************************************************/    
public class Path extends DisplayObject
{
	Node head;
	Node[] nodes;

	// This is so each node can keep a seperate xy for each path throught
	// it so that they will not seem to join and the point in question
	int[] which_point;
	
	int direction = DIR_ONE;
	
	public Path(int direction_input, int num_nodes, Node[] nodes_input)
	{
		size = 1;
		direction = direction_input;
		nodes = new Node[num_nodes];
		which_point = new int[num_nodes];
		if (num_nodes < 1)
			return;
		Node last = null;
		Node current = null;
		Node next = nodes_input[0];
		int i =0;
		while (i < num_nodes)
		{
			last = current;
			current = next;
			if (i < num_nodes-1)
				next = nodes_input[i+1];
			else
				next = null;
			nodes[i] = current;
			which_point[i] = current.addPath(last,next,this);
			i++;
		}
	}

	public void setSelect(boolean input)
	{
		if (input)
			size++;
		else
			size--;
		super.setSelect(input);
	}
		

	public double near(int x, int y)
	{
	   if (!exists || hidden || size < 1)
		return -1;
	   for (int i=0; i< nodes.length-1;i++)
	   {
		Node parent = nodes[i];
		Node child = nodes[i+1];
		int which_parent = which_point[i];
		int which_child = which_point[i+1];
		int pX = parent.getPathX(which_parent);
		int pY = parent.getPathY(which_parent);
		int cX = child.getPathX(which_child);
		int cY = child.getPathY(which_child);
            	int temp;
 
                int low_x = pX;
                int high_x = cX;
                if (low_x > high_x)
                {
                        temp = low_x;
                        low_x = high_x;
                        high_x = temp;
                }
                int low_y = pY;
                int high_y = cY;
                if (low_y > high_y)
                {
                        temp = low_y;
                        low_y = high_y;
                        high_y = temp;
                }
 
                if (low_x <= x && x <= high_x && low_y <= y && y <= high_y)
                {
// REALLY STUPID HACK NO CLUE WHY IT IS NEEDED
x -= pX;
cX -= pX;
pX -= pX;
                        double Xn = pY - cY;
                        double Yn = cX - pX;
                      //jae  thinks it's a bug
                      //double Z = (y*Yn + x*Xn - pY*Yn - pX*Yn);
                        double Z = (y*Yn + x*Xn - pY*Yn - pX*Xn);
                        Z = (Z*Z)/(Xn*Xn + Yn*Yn);
                        if (Z < (size+2)*(2+size))
                                return Z;
                }
	  }
          return -1;
	}

    public int getMinX()
    {
	final int MIN_X=0;
	final int MIN_Y=1;
	final int MAX_X=2;
	final int MAX_Y=3;
	final int MAX_INT=2147483647;

	int[] boundaries=getXYBoundaries();
	if (boundaries==null)
	    return 0;
	else return boundaries[MIN_X];
    }

    public int getMaxX()
    {
	final int MIN_X=0;
	final int MIN_Y=1;
	final int MAX_X=2;
	final int MAX_Y=3;
	final int MAX_INT=2147483647;

	int[] boundaries=getXYBoundaries();
	if (boundaries==null)
	    return 0;
	else return boundaries[MAX_X];
    }


    public int getMinY()
    {
	final int MIN_X=0;
	final int MIN_Y=1;
	final int MAX_X=2;
	final int MAX_Y=3;
	final int MAX_INT=2147483647;

	int[] boundaries=getXYBoundaries();
	if (boundaries==null)
	    return 0;
	else return boundaries[MIN_Y];
    }

    public int getMaxY()
    {
	final int MIN_X=0;
	final int MIN_Y=1;
	final int MAX_X=2;
	final int MAX_Y=3;
	final int MAX_INT=2147483647;

	int[] boundaries=getXYBoundaries();
	if (boundaries==null)
	    return 0;
	else return boundaries[MAX_Y];
    }

    /**Gets the Max and Min x/y values for this path*/
    private int[] getXYBoundaries()
    {
	int[] retval=new int[4];
	
	final int MIN_X=0;
	final int MIN_Y=1;
	final int MAX_X=2;
	final int MAX_Y=3;
	final int MAX_INT=2147483647;
	retval[MIN_Y]=retval[MIN_X]=MAX_INT; //cheez
	retval[MAX_X]=retval[MAX_Y]=-MAX_INT-1;


	if (nodes.length==0){
	    return null;
	}

	for (int i=0; i<nodes.length; i++){
	    Node parent=nodes[i];
	    Node child=nodes[i+1];
	    int which_parent = which_point[i];
	    int which_child = which_point[i+1];
	    int x1 = parent.getPathX(which_parent);
	    int y1 = parent.getPathY(which_parent);
	    int x2 = child.getPathX(which_child);
	    int y2 = child.getPathY(which_child);
	    if (x1<x2){
		if (x1<retval[MIN_X]){
		    retval[MIN_X]=x1;
		}
		if (x2>retval[MAX_X]){
		    retval[MAX_X]=x2;
		}
	    }else {//if (x2<x1){
		if (x2<retval[MIN_X]){
		    retval[MIN_X]=x2;
		}
		if (x1>retval[MAX_X]){
		    retval[MAX_X]=x1;
		}
	    }

	    if (y1<y2){
		if (y1<retval[MIN_Y]){
		    retval[MIN_Y]=y1;
		}
		if (y2>retval[MAX_Y]){
		    retval[MAX_Y]=y2;
		}
	    }else{// if (y2<y1){
		if (y2<retval[MIN_Y]){
		    retval[MIN_Y]=y2;
		}
		if (y1>retval[MAX_Y]){
		    retval[MAX_Y]=y1;
		}
	    }
	}
    
	return retval;
    }

    public void draw(Graphics g, int x_shift, int y_shift, 
	     int x_scale, int y_scale, Dimension dim, double[] latlon,
	     boolean drawing_postscript)
    {
    draw(g,(double)x_shift,(double)y_shift,(double)x_scale,
         (double)y_scale,dim,latlon,drawing_postscript);
    }

   public void draw(Graphics g, double x_shift, double y_shift, 
                     double x_scale, double y_scale, Dimension dim, 
		    double[] latlon, boolean drawing_postscript)
    {

	if (hidden || size < 1)
	    return;
	if (selected)
	    g.setColor(selected_color);
	else
	    g.setColor(color);

        if (!drawing_postscript) {
            x_shift += ShiftImage.getXShift();
            y_shift += ShiftImage.getYShift();
        }

	for (int i=0; i < nodes.length-1;i++)
	{
	    Node parent = nodes[i];
	    Node child = nodes[i+1];
	    int which_parent = which_point[i];
	    int which_child = which_point[i+1];
	    double x1 = parent.getPathX(which_parent);
	    double y1 = parent.getPathY(which_parent);
	    double x2 = child.getPathX(which_child);
	    double y2 = child.getPathY(which_child);

	    // From DisplayObject shared with Link
	    drawLine(g,x1,y1,x2,y2, x_shift,y_shift,x_scale,y_scale,
		drawing_postscript);
       }
    }
	
	public int getDirection()
	{
		return direction;
	}
		
	public String getInfo()
	{
		String middle = "->";
		if (direction == DIR_TWO)
			middle = "<->";
		String answer = "";
		if (current_group != null && current_values != null)
			if (current_group.getType() == ValuesGroup.DOUBLE)
				answer = "(" + double_value +") ";
			else
				answer = "(" + string_value +") ";
		int i =0;
		for (i=0; i < nodes.length-1;i++)
			answer = answer + nodes[i].getName() + middle;
		if (i < nodes.length)
			answer = answer + nodes[i].getName();
		return answer;
	}
}
