package caida.tools;
import java.awt.*;

/**************************************************************************
* File: Strings.java	Class: Strings
* Goal: A array of strings which will grow and can sort itself.
*
***************************************************************************
***************************************************************************
*  By accessing this software, STRINGS, you are duly informed of
*  and agree to be bound by the conditions described below in
*  this notice:
*
*  This software product, STRINGS, is developed by Bradley Huffaker, and
*  copyrighted(C) 1998 by the University of California, San Diego (UCSD),
*  with all rights reserved.  UCSD administers the NSF grant to CAIDA,
*  number NCR-9711092, under which this code was developed. 
*
*  STRINGS is a free software. You can redistribute it and/or modify it
*  under the terms of the GNU Public License v.2 as
*  published by UCSD and is incorporated by reference herein.  STRINGS is
*  distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF MERCHANTABILITY
*  OR FITNESS FOR A PARTICULAR PURPOSE or that the use of it will not
*  infringe on any third party's intellectual property rights.
*
*  You should have received a copy of the CAIDA General Public
*  License v.1 along with the STRINGS program.  If for any reason you
*  have not received a copy, please write to
*
*                       University of California, San Diego
*                       SDSC/CAIDA
*                       9500 Gilman Dr., MS-0505
*                       La Jolla, CA 92093 - 0505  USA
*
*  Or contact INFO@CAIDA.ORG
**************************************************************************/    
public class Strings 
{
	String[] strings;
	int num_strings;
	int[] occurances;

	public Strings(int size)
	{
		strings = new String[size];
                occurances = new int[size];
		num_strings = 0;
	}

	public Strings()
	{
		this(10);
	}

	public int length() { return num_strings; }

	public void set(String string, int index)
	{
		if (index >= num_strings)
			index = num_strings-1;
		strings[index] = string;
                occurances[index] = 1;
	}
	
	public int add(String string, int string_occurances)
	{
		if (num_strings == strings.length)
		{
			int length = (int)(strings.length*1.1);
			String[] temp = new String[length];
			int[] temp_int = new int[length];
                       
			for(int i=0;i < strings.length; i++)
 			{
				temp[i] = strings[i];
				temp_int[i] = occurances[i];
			}
			strings = temp;
			occurances = temp_int;
		}
		strings[num_strings] = string;
		occurances[num_strings] = string_occurances;
		return num_strings++;
	}
	public int add(String string)
	{
		return add(string, 1);
	}
	
	public int addUnique(String string)
	{
		int index = 0;
		while (index < num_strings)
		{
			if (strings[index].equals(string))
			{
				occurances[index]++;
				return index;
			}
			index++;
		}
		return add(string);
	}

	public int remove(String string)
	{
		int index = 0;
		while (index < num_strings && !strings[index].equals(string))
			index++;
		if (index <num_strings)
		{
			num_strings--;
			strings[index] = strings[num_strings];
			occurances[index] = occurances[num_strings];
			return index;
		}
		return -1;
	}		

	public int indexOf(String string)
	{
		int index=0;
		while (index < num_strings)
		{
			if (strings[index].equals(string))
				return index;
			index++;
		}
		return -1;
	}
	
	public String get(int index)
	{
		if (index >= num_strings)
			index = num_strings-1;
		return strings[index];
	}
	
	public int getOccurances(int index)
	{
		if (index >= num_strings)
			index = num_strings-1;
		return occurances[index];
	}

	public void sort()
	{
		for (int index=0; index < num_strings; index++)
		{
			String index_string = strings[index];
			for (int x=num_strings-1; x > index; x--)
			{
				String x_string = strings[x];
				if (less(x_string,index_string))
				{
					strings[x] = index_string;
					index_string = x_string;
					int temp_int = occurances[index];
					occurances[index] = occurances[x];
					occurances[x] = temp_int;
				}
			}
			strings[index] = index_string;
		}
	}
	
	public boolean less(String left, String right)
	{
		char[] l_array = left.toCharArray();
		char[] r_array = right.toCharArray();
		int index=0;
		int shortest = l_array.length;
		if (shortest > r_array.length)
			shortest = r_array.length;
		while (index < shortest)
		{
			if (l_array[index] < r_array[index])
				return true;
			else if (l_array[index] > r_array[index])
				return false;
			index++;
		}
		if (shortest == r_array.length)
			return false;
		return true;
	}
}
